import os
import re
import yaml

from distutils.spawn import find_executable

from popper.cli import log


def sanitized_name(name, wid=""):
    """Clean a step name and change it to proper format. It replaces all the
    unwanted characters with `_`.

    Args:
      name(str): The crud step name.
      wid(str): It  is a workflow ID produced by a utils.get_id().

    Returns:
      str: The sanitize step name.
    """
    return f"popper_{re.sub('[^a-zA-Z0-9_.-]', '_', name)}_{wid}"


def assert_executable_exists(command):
    """Check if the given command can be invoked; fails if not."""
    if not find_executable(command):
        log.fail(f"Could not find '{command}'.")


def prettystr(a):
    """improve how dictionaries get printed"""
    if isinstance(a, os._Environ):
        a = dict(a)
    if isinstance(a, dict):
        return f"{yaml.dump(a, default_flow_style=False)}"


def key_value_to_flag(k, v, equals_symbol=False):
    is_bool = isinstance(v, bool)

    if is_bool and not v and not equals_symbol:
        return ""

    flag = "-" if len(k) == 1 else "--"

    if equals_symbol:
        flag += f"{k}={str(v).lower() if is_bool else v}"
    else:
        if isinstance(v, bool):
            flag += f"{k}"
        else:
            flag += f"{k} {v}"
    return flag



from __future__ import unicode_literals

import logging
import os
import re
import yaml

from popper.cli import log as log

from box import Box

from pykwalify.core import Core as YMLValidator
from pykwalify.errors import SchemaError


class WorkflowParser(object):
    _wf_schema = {
        "type": "map",
        "mapping": {
            "steps": {
                "type": "seq",
                "sequence": [
                    {
                        "type": "map",
                        "mapping": {
                            "uses": {"type": "str", "required": True},
                            "id": {"type": "str"},
                            "args": {"type": "seq", "sequence": [{"type": "str"}]},
                            "runs": {"type": "seq", "sequence": [{"type": "str"}]},
                            "secrets": {"type": "seq", "sequence": [{"type": "str"}]},
                            "dir": {"type": "str"},
                            "skip_pull": {"type": "bool"},
                            "env": {
                                "type": "map",
                                "matching-rule": "any",
                                "mapping": {"regex;(.+)": {"type": "str"}},
                            },
                            "options": {"type": "map", "allowempty": True},
                        },
                    }
                ],
            },
            "options": {
                "type": "map",
                "mapping": {
                    "secrets": {"type": "seq", "sequence": [{"type": "str"}]},
                    "env": {
                        "type": "map",
                        "matching-rule": "any",
                        "mapping": {"regex;(.+)": {"type": "str"}},
                    },
                },
            },
        },
    }

    @staticmethod
    def parse(
        file=None,
        wf_data=None,
        step=None,
        skipped_steps=[],
        substitutions=[],
        allow_loose=False,
        immutable=True,
    ):
        """Returns an immutable workflow structure (a frozen Box) with 'steps' and
        'options' properties. See WorkflowParser._wf_schema above for their structure.
        If immutable=False is given, the returned object representing the workflow
        can be modified (an "unfrozen" ``box.Box``).
        """

        if not file and not wf_data:
            log.fail("Expecting 'file' or 'wf_data'")

        if file:
            if wf_data:
                log.fail("Expecting only one of 'file' and 'wf_data'")

            if not os.path.exists(file):
                log.fail(f"File {file} was not found.")

            if not file.endswith(".yml") and not file.endswith(".yaml"):
                log.fail("Unrecognized workflow file format.")

            with open(file, "r") as f:
                _wf_data = yaml.safe_load(f)

                if not _wf_data:
                    log.fail(f"File {file} is empty")
        else:
            _wf_data = dict(wf_data)

        # disable logging in order to silence warnings about "error() to fail()" change
        logging.disable(logging.CRITICAL)

        v = YMLValidator(source_data=_wf_data, schema_data=WorkflowParser._wf_schema)

        try:
            v.validate()
        except SchemaError as e:
            # reenable logging
            logging.disable(logging.NOTSET)
            log.fail(f"{e.msg}")

        logging.disable(logging.NOTSET)

        WorkflowParser.__add_missing_ids(_wf_data)
        WorkflowParser.__skip_steps(_wf_data, skipped_steps)
        WorkflowParser.__filter_step(_wf_data, step)
        WorkflowParser.__propagate_options_to_steps(_wf_data)
        WorkflowParser.__apply_substitutions(
            _wf_data, substitutions=substitutions, allow_loose=allow_loose
        )

        # create and frozen a box
        wf_box = Box(_wf_data, frozen_box=(immutable is True), default_box=True)

        log.debug(f"Parsed workflow:\n{wf_box}")

        return wf_box

    @staticmethod
    def __apply_substitution(wf_element, k, v, used_registry):
        if isinstance(wf_element, str):
            if k in wf_element:
                log.debug(f"Applying substitution to string {k}")
                wf_element = wf_element.replace(k, v)
                used_registry[k] = 1

        elif isinstance(wf_element, list):
            # we assume list of strings
            for i, e in enumerate(wf_element):
                if k in e:
                    log.debug(f"Applying substitution to item {i}: {e}")
                    wf_element[i] = wf_element[i].replace(k, v)
                    used_registry[k] = 1

        elif isinstance(wf_element, dict):
            # we assume map of strings
            for ek in wf_element:
                if k in ek:
                    log.fail("Substitutions not allowed on dictionary keys")
                if k in wf_element[ek]:
                    log.debug(f"Applying substitution to value associated to key {k}")
                    wf_element[ek] = wf_element[ek].replace(k, v)
                    used_registry[k] = 1

        return wf_element

    @staticmethod
    def __add_missing_ids(wf_data):
        for i, step in enumerate(wf_data["steps"]):
            step["id"] = step.get("id", f"{i+1}")

    @staticmethod
    def __propagate_options_to_steps(wf_data):
        """Copies env and secrets attributes from 'options' to each step. Step
        attributes have precedence over workflow-wide ones
        """
        # we create dict/list with env/secrets from 'options'
        wf_env = wf_data.get("options", {}).get("env", {})
        wf_secrets = wf_data.get("options", {}).get("secrets", [])

        # for each step, create a copy of the above, and update with info from step in
        # order to make step have higher precedence over workflow-wide options
        for step in wf_data["steps"]:
            step_env = dict(wf_env)
            step_env.update(step.get("env", {}))
            step["env"] = step_env

            step_secrets = wf_secrets + step.get("secrets", [])
            step["secrets"] = step_secrets

    @staticmethod
    def __apply_substitutions(wf_data, substitutions=None, allow_loose=False):
        if not substitutions:
            return

        used = {}
        for substitution in substitutions:
            item = substitution.split("=", 1)
            if len(item) < 2:
                raise Exception("Excepting '=' as seperator")

            k, v = (item[0], item[1])

            if not re.match(r"_[A-Z0-9]+", k):
                log.fail(f"Expecting substitution key in _[A-Z0-9] format; got '{k}'.")

            # add dollar sign
            k = "$" + k

            # replace in steps
            for step in wf_data["steps"]:
                for attr in step:
                    step[attr] = WorkflowParser.__apply_substitution(
                        step[attr], k, v, used
                    )
            opts = wf_data.get("options", {})
            for attr in opts:
                opts[attr] = WorkflowParser.__apply_substitution(opts[attr], k, v, used)

        if not allow_loose and len(substitutions) != len(used):
            log.fail("Not all given substitutions are used in the workflow file")

        log.debug(f"Workflow after applying substitutions: {wf_data}")

    @staticmethod
    def __skip_steps(wf_data, skip_list=[]):
        if not skip_list:
            return
        filtered_list = []
        used = {}
        for step in wf_data["steps"]:
            if step["id"] in skip_list:
                used[step["id"]] = 1
                continue
            filtered_list.append(step)
        wf_data["steps"] = filtered_list

        if len(used) != len(skip_list):
            log.fail("Not all skipped steps exist in the workflow.")

    @staticmethod
    def __filter_step(wf_data, filtered_step=None):
        """Remove all but the given one."""
        if not filtered_step:
            return
        for step in wf_data["steps"]:
            if step["id"] == filtered_step:
                wf_data["steps"] = [step]
                return
        log.fail(f"Step {filtered_step} not in workflow")



import click
import os
import traceback

from popper import log as logging
from popper.cli import log, pass_context
from popper.config import ConfigLoader
from popper.parser import WorkflowParser
from popper.runner import WorkflowRunner


@click.command("run", short_help="Run a workflow or step.")
@click.argument("step", required=False)
@click.option(
    "-f",
    "--file",
    help="File containing the definition of the workflow.",
    required=False,
    default=".popper.yml",
)
@click.option(
    "-d",
    "--debug",
    help="Generate detailed messages of what popper does (overrides --quiet)",
    required=False,
    is_flag=True,
)
@click.option(
    "--dry-run",
    help="Do not run the workflow, only print what would be executed.",
    required=False,
    is_flag=True,
)
@click.option(
    "--log-file",
    help="Path to a log file. No log is created if this is not given.",
    required=False,
)
@click.option(
    "-q",
    "--quiet",
    help="Do not print output generated by steps.",
    required=False,
    is_flag=True,
)
@click.option(
    "--reuse",
    help="Reuse containers between executions (persist container state).",
    required=False,
    is_flag=True,
)
@click.option(
    "-e",
    "--engine",
    help="Specify container engine used to execute workflow steps.",
    type=click.Choice(["docker", "singularity", "podman"]),
)
@click.option(
    "-r",
    "--resource-manager",
    help="Specify resource manager for executing the workflow.",
    type=click.Choice(["host", "slurm", "kubernetes"]),
)
@click.option(
    "--skip",
    help="Skip the given step (can be given multiple times).",
    required=False,
    default=list(),
    multiple=True,
)
@click.option(
    "--skip-clone",
    help="Skip cloning repositories (assume they have been cloned).",
    required=False,
    is_flag=True,
)
@click.option(
    "--skip-pull",
    help="Skip pulling container images (assume they exist in local cache).",
    required=False,
    is_flag=True,
)
@click.option(
    "-s",
    "--substitution",
    help="A key-value pair defining a substitution. " "Can be given multiple times.",
    required=False,
    default=list(),
    multiple=True,
)
@click.option(
    "--allow-loose",
    help=(
        "Do not throw an error if a substitution variable passed as an "
        "argument is unused in the workflow definition."
    ),
    required=False,
    is_flag=True,
)
@click.option(
    "--allow-undefined-secrets-in-ci",
    help=(
        "Do not throw an error if a secret is undefined and the CI environment"
        "variable is defined."
    ),
    required=False,
    is_flag=True,
)
@click.option(
    "-w",
    "--workspace",
    help="Path to workspace folder.",
    required=False,
    show_default=False,
    default=os.getcwd(),
)
@click.option(
    "-c", "--conf", help="Path to file with configuration options.", required=False
)
@pass_context
def cli(
    ctx,
    step,
    file,
    debug,
    dry_run,
    log_file,
    quiet,
    reuse,
    engine,
    resource_manager,
    skip,
    skip_pull,
    skip_clone,
    substitution,
    allow_loose,
    allow_undefined_secrets_in_ci,
    workspace,
    conf,
):
    """Runs a Popper workflow. Only executes STEP if given.

    To specify a container engine to use other than docker, use the --engine/-e
    flag. For executing on a resource manager such as SLURM or Kubernetes, use
    the --resource-manager/-r flag. Alternatively, a configuration file can be
    given (--conf flag) that can specify container options, resource manager
    options, or both (see "Workflow Syntax and Execution Runtime" section of
    the Popper documentation for more).

    If the container engine (-e) or resource manager (-r) are specified with a
    flag and a configuration file is given as well, the values passed via the
    flags are given preference over those contained in the configuration file.
    """
    # set the logging levels.
    level = "STEP_INFO"
    if quiet:
        level = "INFO"
    if debug:
        level = "DEBUG"
    log.setLevel(level)

    if dry_run:
        logging.msg_prefix = "DRYRUN: "

    if log_file:
        # also log to a file
        logging.add_log(log, log_file)

    # check conflicting flags and fail if needed
    if skip and step:
        log.fail("`--skip` can not be used when STEP argument is passed.")

    # invoke wf factory; handles formats, validations, filtering
    wf = WorkflowParser.parse(
        file,
        step=step,
        skipped_steps=skip,
        substitutions=substitution,
        allow_loose=allow_loose,
    )

    config = ConfigLoader.load(
        engine_name=engine,
        resman_name=resource_manager,
        config_file=conf,
        reuse=reuse,
        dry_run=dry_run,
        skip_pull=skip_pull,
        skip_clone=skip_clone,
        workspace_dir=workspace,
        allow_undefined_secrets_in_ci=allow_undefined_secrets_in_ci,
    )

    with WorkflowRunner(config) as runner:
        try:
            runner.run(wf)
        except Exception as e:
            log.debug(traceback.format_exc())
            log.fail(e)



import os
import re

import git

from popper.cli import log


def new_repo(gitrepo_dir=None):
    """Function to initialize a git.Repo object for a repository that is
    assumed to be in gitrepo_dir, or any parent directory.

    Args:
        gitrepo_dir(str): path to a folder within a git repository folder

    Returns:
        git.Repo: repo object or None if '.git/' not found in any parent folder
    """
    if not gitrepo_dir or not os.path.isdir(gitrepo_dir):
        return None

    repo = None

    try:
        repo = git.Repo(gitrepo_dir, search_parent_directories=True)
    except git.InvalidGitRepositoryError:
        # Optimistically assume that this is due to .git/ folder not existing
        pass

    if not repo or is_empty(repo):
        return None

    return repo


def get_remote_url(repo=None):
    """Obtains remote origin URL, if possible. Otherwise it returns empty str.

    Args:
        None

    Returns:
      str: The remote origin url or "".
    """
    url = ""
    if repo:
        if len(repo.remotes) > 0:
            url = repo.remotes.origin.url
        # cleanup the URL so we get in in https form and without '.git' ending
        if url.endswith(".git"):
            url = url[:-4]
        if "git@" in url:
            url = "https://" + url[4:].replace(":", "/")

    return url


def clone(url, org, repo, repo_dir, version=None):
    """Clones a repository using Git. If ``repo_dir`` already exists, it pulls
    from the remote.

    Args:
      url(str): The remote git repository hosting service url.
      org(str): The org/user to which the repo belongs.
      repo(str): The repo name.
      repo_dir(str): The path where to clone the repo.
      version(str, optional): The remote tag/branch to checkout. If version is
                              None, we use the default remote branch as version

    Returns:
        None

    """

    def get_default_branch(r):
        """
        Used to find default branch of the repository.

        Args:
          r: It is git.Repo object

        Returns:

        """
        if version:
            return version
        return r.remotes.origin.refs["HEAD"].ref.remote_head

    if os.path.exists(repo_dir):
        cloned_repo = git.Repo(repo_dir)
        cloned_repo.remotes.origin.pull(get_default_branch(cloned_repo))
    else:
        if "@" in url:
            url += ":"
        else:
            url += "/"

        # To obtain the authentication token if set as environment variable.
        auth_token = os.getenv("GITHUB_API_TOKEN")

        if auth_token is not None and "github" in url and "@" not in url:
            # To verify the link of github for private repo support.
            # The authentication token has to be added after protocol
            # The length of protocol is 8 in case of https://
            url = url[:8] + auth_token + "@" + url[8:]

        repo_url = f"{url}{org}/{repo}"
        cloned_repo = git.Repo.clone_from(repo_url, repo_dir)

    cloned_repo.git.checkout(get_default_branch(cloned_repo))
    cloned_repo.close()


def parse(url):
    """Method to parse the git url. If a URL schema and hostname are not 
    included, it assumes ``https://github.com``.

    Args:
      url(str): The url in string format.

    Returns:
        tuple(service_url, service, user, repo, step_dir, version)

    """

    if url.startswith("ssh://"):
        log.fail("The ssh protocol is not supported yet.")

    if url.endswith(".git"):
        url = url[:-4]

    pattern = re.compile(
        r"^(http://|https://|git@)?(?:(\w+\.\w+)(?:\/|\:))?"
        r"([\w\-]+)(?:\/([^\@^\/]+)\/?([^\@]+)?(?:\@([\w\W]+))?)$"
    )

    try:
        protocol, service, user, repo, step_dir, version = pattern.search(url).groups()
    except AttributeError:
        log.fail(
            "Invalid url. The url should be in any of the 3 forms: \n"
            "1) https://github.com/user/repo/path/to/step@version \n"
            "2) gitlab.com/user/repo/path/to/step@version \n"
            "3) user/repo/path/to/step@version"
        )

    if not service:
        service = "github.com"

    if not protocol:
        protocol = "https://"

    if not step_dir:
        step_dir = ""

    service_url = protocol + service

    log.debug(f'parse("{url}"):')
    log.debug(f"  service_url: {service_url}")
    log.debug(f"  service: {service}")
    log.debug(f"  user: {user}")
    log.debug(f"  repo: {repo}")
    log.debug(f"  step_dir: {step_dir}")
    log.debug(f"  version: {version}")

    return service_url, service, user, repo, step_dir, version


def is_empty(repo):
    """True if the currently checked out branch has no commits."""
    return repo.git.rev_list("-n 1", "--all") == ""


def get_sha(repo, short=None):
    """Returns the commit id for the currently checked out version on the
    given repository object. If short is given, it is interpreted as the number
    of characters from the SHA that get returned. E.g. short=7 returns the
    first 7 characters, otherwise it returns the entire SHA1 string.
    """
    if not repo or is_empty(repo):
        return None

    if short:
        sha = repo.git.rev_parse(repo.head.object.hexsha, short=short)
    else:
        sha = repo.git.rev_parse(repo.head.object.hexsha)

    return sha


def get_branch(repo):
    """Get name of branch. If the repo is in detached head state, it looks for
    for environment variables commonly used in CI services: TRAVIS_BRANCH,
    GIT_BRANCH (Jenkins), CIRCLE_BRANCH and CI_COMMIT_REF_NAME (Gitlab)
    """
    if not repo or is_empty(repo):
        return None

    if not repo.head.is_detached:
        return repo.active_branch.name

    for v in ["TRAVIS_BRANCH", "GIT_BRANCH", "CIRCLE_BRANCH", "CI_COMMIT_REF_NAME"]:
        branch = os.environ.get(v)
        if branch:
            return branch

    return get_sha(repo)


def get_tag(repo):
    """Get name of tag for current commit. If current commit is not tagged,
    it looks for for environment variables commonly used in CI services:
    TRAVIS_TAG, CIRCLE_TAG, GIT_TAG (Jenkins) and CI_COMMIT_REF_NAME (Gitlab).
    """
    if not repo or is_empty(repo):
        return None

    tags = repo.git.tag("--points-at", "HEAD").split()

    if len(tags) > 0:
        return tags[0]

    for v in ["TRAVIS_TAG", "GIT_TAG", "CIRCLE_TAG", "CI_COMMIT_REF_NAME"]:
        tag = os.environ.get(v)
        if tag:
            return tag

    return None



from click.testing import CliRunner

from popper import __version__
from popper import _version_file
from popper.commands import cmd_version
from .test_common import PopperTest


class TestCommandVersion(PopperTest):
    def test_version(self):
        self.assertIsNot("0.0.0", __version__)
        with self.assertLogs("popper") as test:
            result = CliRunner().invoke(cmd_version.cli)
            self.assertTrue(__version__ in test.output[0])
            self.assertEqual(0, result.exit_code)

        with open(_version_file) as f:
            self.assertEqual(f"__popper_version__ = '{__version__}'\n", f.read())



import os

import click

from popper.cli import pass_context, log


@click.command("scaffold", short_help="Generate a minimal workflow.")
@click.option(
    "-f",
    "--file",
    help="Name of file where to write the generated workflow.",
    required=False,
    default="wf.yml",
)
@pass_context
def cli(ctx, file):
    """Generates a minimal workflow that can be used as starting point."""
    main_workflow_content = """steps:

- uses: "popperized/bin/sh@master"
  args: ["ls"]

- uses: "docker://alpine:3.11"
  args: ["ls"]
"""

    if os.path.exists(file):
        log.fail(f"File {file} already exists")

    with open(file, "w") as f:
        f.write(main_workflow_content)

    log.info("Successfully generated a workflow scaffold.")



import logging
import sys
import os

STEP_INFO = 15
logging.addLevelName(STEP_INFO, "STEP_INFO")

msg_prefix = ""


class PopperFormatter(logging.Formatter):
    """A Formatter which sets the format for popper custom log messages.

    Level Values are
    DEBUG: 10
    STEP_INFO: 15
    INFO: 20
    WARNING: 30
    ERROR: 40
    CRITICAL: 50

    The level STEP_INFO is used to log information produced solely by
    steps during popper execution.
    The popper default log level is STEP_INFO. On the usage of --quiet flag,
    we change the level to INFO, thus effectively silencing STEP_INFO log.

    In order of Level. The format of the logs is given in log_format dict.
    The log colors used are based on ANSI Escape Codes
    """

    # Log Colors
    BOLD_CYAN = "[01;36m"
    RESET = "[0m"
    BOLD_YELLOW = "[01;33m"
    BOLD_RED = "[01;31m"

    log_format = {
        "DEBUG": f"{BOLD_CYAN}%(levelname)s: %(msg)s {RESET}",
        "STEP_INFO": "%(msg)s",
        "INFO": "%(msg)s",
        "WARNING": f"{BOLD_YELLOW}%(levelname)s: %(msg)s{RESET}",
        "ERROR": f"{BOLD_RED}%(levelname)s: %(msg)s{RESET}",
        "CRITICAL": f"{BOLD_RED}%(levelname)s: %(msg)s{RESET}",
    }

    log_format_no_colors = {
        "DEBUG": "%(levelname)s: %(msg)s ",
        "STEP_INFO": "%(msg)s",
        "INFO": "%(msg)s",
        "WARNING": "%(levelname)s: %(msg)s",
        "ERROR": "%(levelname)s: %(msg)s",
        "CRITICAL": "%(levelname)s: %(msg)s",
    }

    def __init__(self, colors=True):
        super(PopperFormatter, self).__init__(fmt="%(levelname)s: %(msg)s")
        self.log_fmt = self.log_format if colors else self.log_format_no_colors

    def format(self, record):
        """

        Args:
          record(logging.LogRecord): The part of the log record from which
                                        the information is to be extracted.

        Returns:
          str: String containing meaningful information from logs.

        """
        fmt = self.log_fmt[logging.getLevelName(record.levelno)]
        if sys.version_info[0] < 3:
            self._fmt = fmt
        else:
            self._style._fmt = fmt
        result = f"{msg_prefix}{logging.Formatter.format(self, record)}"
        return result


class PopperLogger(logging.Logger):
    """A Logger so that we can add popper fail and step_info log methods."""

    def fail(self, msg="", *args, **kwargs):
        """Log a message with severity 'ERROR', and then exits.

        Args:
          msg(str, optional): Message to be logged.(Default value = '')
          *args(list): List of non-key worded,variable length arguments.
          **kwargs(dict): List of key-worded,variable length arguments.

        Returns:
          None
        """
        super(PopperLogger, self).error(msg, *args, **kwargs)
        sys.exit(1)

    def step_info(self, msg="", *args, **kwargs):
        """Log a message with severity 'STEP_INFO'.

        Args:
          msg(str, optional): Message to be logged.(Default value = '')
          *args(list): List of non-key worded,variable length arguments.
          **kwargs(dict): List of key-worded,variable length arguments.

        Returns:
          None
        """
        if self.isEnabledFor(STEP_INFO):
            self._log(STEP_INFO, msg, args, **kwargs)

    def error(self, msg, *args, **kwargs):
        """Sends a warning about the use of fail(). We are depreciating it in
        favor of fail()

        The error() method has been replaced with fail(), as fail is more
        indicative of the functionality provided i.e. fail() will also
        result in the failure of the module when called and stop execution.

        Args:
          msg(str, optional): Message to be logged.
          *args(list): List of non-key worded,variable length arguments.
          **kwargs(dict): List of key-worded,variable length arguments.

        Returns:
          None
        """
        super(PopperLogger, self).warning("error() has been replaced with fail()")
        pass

    def info(self, msg="", *args, **kwargs):
        """Logs a message with severity 'INFO'.

        Args:
          msg(str, optional): Message to be logged.(Default value = '')
          *args(list): List of non-key worded,variable length arguments.
          **kwargs(dict): List of key-worded,variable length arguments.

        Returns:
          None
        """
        super(PopperLogger, self).info(msg, *args, **kwargs)

    def debug(self, msg="", *args, **kwargs):
        """Logs a message with severity 'DEBUG'.

        Args:
          msg(str,optional): Message to be logged.(Default value = '')
          *args(list): List of non-key worded,variable length arguments.
          **kwargs(dict): List of key-worded,variable length arguments.

        Returns:
          None
        """
        super(PopperLogger, self).debug(msg, *args, **kwargs)

    def warning(self, msg="", *args, **kwargs):
        """Logs a message with severity 'WARNING'.

        Args:
          msg(str, optional): Message to be logged.(Default value = '')
          *args(list): List of non-key worded,variable length arguments.
          **kwargs(dict): List of key-worded,variable length arguments.

        Returns:
          None
        """
        super(PopperLogger, self).warning(msg, *args, **kwargs)


class LevelFilter(logging.Filter):
    """Filters the level that are to be accepted and rejected."""

    def __init__(self, passlevels, reject):
        self.passlevels = passlevels
        self.reject = reject

    def filter(self, record):
        """Returns True and False according to the pass levels and reject value.

        Args:
          record(logging.LogRecord): Record from logs.

        Returns:
          bool : True/False according to values of pass levels and level number
                of the record.
        """
        if self.reject:
            return record.levelno not in self.passlevels
        else:
            return record.levelno in self.passlevels


def setup_logging(level="STEP_INFO"):
    """Setups logging facilities with custom Logger and Formatter.

    Args:
      level(str): Level to be logged in custom logger.
                    (Default value = 'STEP_INFO')

    Returns:
      popper.log.PopperLogger: Custom log for that particular level.
    """
    logging.setLoggerClass(PopperLogger)
    log = logging.getLogger("popper")

    formatter = PopperFormatter()

    # INFO/STEP_INFO goes to stdout
    h1 = logging.StreamHandler(sys.stdout)
    h1.addFilter(LevelFilter([logging.INFO, STEP_INFO], False))
    h1.setFormatter(formatter)

    # anything goes to stdout
    h2 = logging.StreamHandler(sys.stderr)
    h2.addFilter(LevelFilter([logging.INFO, STEP_INFO], True))
    h2.setFormatter(formatter)

    log.addHandler(h1)
    log.addHandler(h2)
    log.setLevel(level)

    return log


def add_log(log, logfile):
    """It sets the formatter for the handle and add that handler to the logger.

    Args:
      log(Logging.logger): The logger object used for logging.
      logfile(str): path for the log file.

    Returns:
      None
    """
    dir = os.path.dirname(logfile)
    if not os.path.exists(dir) and dir != "":
        os.makedirs(dir)
    handler = logging.FileHandler(logfile)
    formatter = PopperFormatter(colors=False)

    # Set
    handler.setFormatter(formatter)
    log.addHandler(handler)



import os
import signal
import threading

import docker
import dockerpty

from subprocess import Popen, STDOUT, PIPE, SubprocessError, CalledProcessError

import spython
from spython.main.parse.parsers import DockerParser
from spython.main.parse.writers import SingularityWriter

from popper import utils as pu
from popper.cli import log as log
from popper.runner import StepRunner as StepRunner


class HostRunner(StepRunner):
    """Run a step directly on the host machine."""

    def __init__(self, **kw):
        super(HostRunner, self).__init__(**kw)

        self._spawned_pids = set()

        if self._config.reuse:
            log.warning("Reuse not supported for HostRunner.")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, traceback):
        pass

    def run(self, step):
        step_env = self._prepare_environment(step, env=dict(os.environ))

        if not step.runs:
            raise AttributeError("Expecting 'runs' attribute in step.")
        cmd = step.runs + tuple(step.args)

        log.info(f"[{step.id}] {cmd}")

        if self._config.dry_run:
            return 0

        log.debug(f"Environment:\n{pu.prettystr(step_env)}")

        pid, ecode, _ = HostRunner._exec_cmd(
            cmd, env=step_env, cwd=self._config.workspace_dir, pids=self._spawned_pids
        )
        if pid != 0:
            self._spawned_pids.remove(pid)

        return ecode

    def stop_running_tasks(self):
        for pid in self._spawned_pids:
            log.info(f"Stopping proces {pid}")
            os.kill(pid, signal.SIGKILL)

    @staticmethod
    def _exec_cmd(cmd, env=None, cwd=os.getcwd(), pids=set(), logging=True):
        pid = 0
        ecode = None
        try:
            with Popen(
                cmd,
                stdout=PIPE,
                stderr=STDOUT,
                universal_newlines=True,
                preexec_fn=os.setsid,
                env=env,
                cwd=cwd,
            ) as p:
                pid = p.pid
                pids.add(p.pid)
                log.debug("Reading process output")

                output = []
                for line in iter(p.stdout.readline, ""):
                    if logging:
                        log.step_info(line.rstrip())
                    else:
                        output.append(line.rstrip())

                p.wait()
                ecode = p.poll()

            log.debug(f"Code returned by process: {ecode}")

        except SubprocessError as ex:
            output = ""
            if not ecode:
                ecode = 1
            log.step_info(f"Command '{cmd[0]}' failed with: {ex}")
        except Exception as ex:
            output = ""
            ecode = 1
            log.step_info(f"Command raised non-SubprocessError error: {ex}")

        return pid, ecode, "\n".join(output)


class DockerRunner(StepRunner):
    """Runs steps in docker on the local machine."""

    def __init__(self, init_docker_client=True, **kw):
        super(DockerRunner, self).__init__(**kw)

        self._spawned_containers = set()
        self._d = None

        if not init_docker_client:
            return

        try:
            self._d = docker.from_env()
            self._d.version()
        except Exception as e:
            log.debug(f"Docker error: {e}")
            log.fail("Unable to connect to the docker daemon.")

        log.debug(f"Docker info: {pu.prettystr(self._d.info())}")

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self._d:
            self._d.close()

    def run(self, step):
        """Execute the given step in docker."""
        cid = pu.sanitized_name(step.id, self._config.wid)

        container = self._find_container(cid)

        if not container and self._config.reuse:
            log.fail(
                f"Cannot find an existing container for step '{step.id}' to be reused"
            )

        if container and not self._config.reuse and not self._config.dry_run:
            container.remove(force=True)
            container = None

        if not container and not self._config.reuse:
            container = self._create_container(cid, step)

        log.info(f"[{step.id}] docker start")

        if self._config.dry_run:
            return 0

        self._spawned_containers.add(container)

        try:
            container.start()

            if self._config.pty:
                dockerpty.start(self._d.api, container.id)
            else:
                cout = container.logs(stream=True)
                for line in cout:
                    log.step_info(line.decode().rstrip())

            e = container.wait()["StatusCode"]
        except Exception as exc:
            log.fail(exc)
        return e

    def stop_running_tasks(self):
        for c in self._spawned_containers:
            log.info(f"Stopping container {c.name}")
            c.stop()

    def _create_container(self, cid, step):
        build, _, img, tag, build_ctx_path = self._get_build_info(step)

        if build:
            log.info(f"[{step.id}] docker build {img}:{tag} {build_ctx_path}")
            if not self._config.dry_run:
                streamer = self._d.api.build(
                    decode=True, path=build_ctx_path, tag=f"{img}:{tag}", rm=True,
                )
                for chunk in streamer:
                    if self._config.quiet:
                        continue
                    if "stream" in chunk:
                        lines = [line for line in chunk["stream"].splitlines() if line]
                        for line in lines:
                            log.step_info(line.strip())

        elif not self._config.skip_pull and not step.skip_pull:
            log.info(f"[{step.id}] docker pull {img}:{tag}")
            if not self._config.dry_run:
                self._d.images.pull(repository=f"{img}:{tag}")

        if self._config.dry_run:
            return

        container_args = self._get_container_kwargs(step, f"{img}:{tag}", cid)

        if "volumes" not in container_args:
            container_args["volumes"] = []
        else:
            container_args["volumes"] = list(container_args["volumes"])

        container_args["volumes"].append("/var/run/docker.sock:/var/run/docker.sock")

        log.debug(f"Container args: {container_args}")

        msg = f"[{step.id}] docker create name={cid}"
        msg += f' image={container_args["image"]}'
        if container_args["entrypoint"]:
            msg += f' entrypoint={container_args["entrypoint"]}'
        if container_args["command"]:
            msg += f' command={container_args["command"]}'
        log.info(msg)

        container = self._d.containers.create(**container_args)

        return container

    def _find_container(self, cid):
        """Check whether the container exists."""
        containers = self._d.containers.list(all=True, filters={"name": cid})

        filtered_containers = [c for c in containers if c.name == cid]

        if len(filtered_containers):
            return filtered_containers[0]

        return None


class PodmanRunner(StepRunner):

    """Runs steps in podman on the local machine."""

    def __init__(self, init_podman_client=True, **kw):
        super(PodmanRunner, self).__init__(**kw)

        self._spawned_containers = set()

        if not init_podman_client:
            return

        try:
            _, _, self._p_info = HostRunner._exec_cmd(["podman", "info"], logging=False)
            self._p_version = HostRunner._exec_cmd(["podman", "version"], logging=False)
        except Exception as e:
            log.debug(f"Podman error: {e}")
            log.fail("Unable to connect to podman, is it installed?")

        log.debug(f"Podman info: {pu.prettystr(self._p_info)}")

    def run(self, step):
        """Executes the given step in podman."""
        cid = pu.sanitized_name(step.id, self._config.wid)
        container = self._find_container(cid)

        if not container and self._config.reuse:
            log.fail(
                f"Cannot find an existing container for step '{step.id}' to be reused"
            )

        if container and not self._config.reuse and not self._config.dry_run:
            cmd = ["podman", "rm", "-f", container]
            HostRunner._exec_cmd(cmd, logging=False)
            container = None

        if not container and not self._config.reuse:
            container = self._create_container(cid, step)

        log.info(f"[{step.id}] podman start")

        if self._config.dry_run:
            return 0

        self._spawned_containers.add(container)

        cmd = ["podman", "start", "-a", container]
        _, e, _ = HostRunner._exec_cmd(cmd)

        return e

    def stop_running_tasks(self):
        """Stop containers started by Popper."""
        for c in self._spawned_containers:
            log.info(f"Stopping container {c}")
            _, ecode, _ = HostRunner._exec_cmd(["podman", "stop", c], logging=False)
            if ecode != 0:
                log.warning(f"Failed to stop the {c} container")

    def _find_container(self, cid):
        """Checks whether the container exists."""
        cmd = ["podman", "inspect", "-f", "{{.Id}}", cid]
        _, ecode, containers = HostRunner._exec_cmd(cmd, logging=False)

        if ecode == 125:
            return None

        if ecode != 0:
            log.fail(f"podman inspect fail: {containers}")

        return containers.strip()

    def _create_container(self, cid, step):
        build, _, img, tag, build_ctx_path = self._get_build_info(step)

        if build:
            log.info(f"[{step.id}] podman build {img}:{tag} {build_ctx_path}")
            if not self._config.dry_run:
                cmd = [
                    "podman",
                    "build",
                    "--tag",
                    f"{img}:{tag}",
                    "--rm",
                    "--file",
                    build_ctx_path,
                ]
                HostRunner._exec_cmd(cmd)
        elif not self._config.skip_pull and not step.skip_pull:
            log.info(f"[{step.id}] podman pull {img}:{tag}")
            if not self._config.dry_run:
                cmd = ["podman", "pull", f"{img}:{tag}"]
                HostRunner._exec_cmd(cmd, logging=False)

        if self._config.dry_run:
            return

        container_args = self._get_container_kwargs(step, f"{img}:{tag}", cid)
        log.debug(f"Container args: {container_args}")

        msg = [f"{step.id}", "podman", "create", f"name={cid}"]
        msg.append(f"image={container_args.get('image')}")
        msg.append(f"entrypoint={container_args.get('entrypoint')}" or "")
        msg.append(f"command={container_args.get('command')}" or "")
        log.info(msg)

        cmd = ["podman", "create"]

        cmd.extend(["--name", container_args.get("name") or ""])
        cmd.extend(["-v", container_args.get("volumes")[0] or ""])

        env = container_args.get("environment")
        if env:
            for i, j in env.items():
                cmd.extend(["-e", f"{i}={j}"])

        cmd.extend(["-d" if container_args.get("detach") else ""])

        cmd.extend(["-w", container_args.get("working_dir")])

        h = container_args.get("hostname", None)
        if h:
            cmd.extend(["-h", h])

        domain_name = container_args.get("domainname", None)
        if domain_name:
            cmd.extend(["--domainname", domain_name])

        tty = container_args.get("tty", None)
        if tty:
            cmd.extend(["-t", tty])

        entrypoint = container_args.get("entrypoint")
        if entrypoint:
            cmd.extend(["--entrypoint", entrypoint[0]])
            entrypoint_rmd = entrypoint[1:]

        cmd.append(container_args.get("image"))

        if entrypoint:
            for i in entrypoint_rmd:
                cmd.extend([i or ""])

        for i in container_args["command"]:
            cmd.extend([i or ""])

        _, ecode, container = HostRunner._exec_cmd(cmd, logging=False)

        if ecode != 0:
            return None

        container = container.rsplit()
        if len(container) == 0:
            return None

        return container[-1]


class SingularityRunner(StepRunner):
    """Runs steps in singularity on the local machine."""

    lock = threading.Lock()

    def __init__(self, init_spython_client=True, **kw):
        super(SingularityRunner, self).__init__(**kw)

        self._spawned_containers = set()
        self._s = None

        if SingularityRunner._in_docker():
            log.fail(
                (
                    "You seem to be running Popper in a Docker container.\n"
                    "Singularity cannot be executed this way.\n"
                    "Either run Popper without Singularity or install Popper "
                    "through PIP.\n"
                    "Instructions are available here:\n"
                    "https://github.com/getpopper/popper/"
                    "blob/master/docs/installation.md"
                )
            )

        if self._config.reuse:
            log.fail("Reuse not supported for SingularityRunner.")

        if not init_spython_client:
            return

        self._s = spython.main.Client
        self._s.quiet = True

    def run(self, step):
        self._setup_singularity_cache()
        cid = pu.sanitized_name(step.id, self._config.wid) + ".sif"
        self._container = os.path.join(self._singularity_cache, cid)

        exists = os.path.exists(self._container)
        if exists and not self._config.dry_run and not self._config.skip_pull:
            os.remove(self._container)

        self._create_container(step, cid)
        ecode = self._singularity_start(step, cid)
        return ecode

    @staticmethod
    def _convert(dockerfile, singularityfile):
        parser = DockerParser(dockerfile)
        for p in parser.recipe.files:
            p[0] = p[0].strip('"')
            p[1] = p[1].strip('"')
            if os.path.isdir(p[0]):
                p[0] += "/."

        writer = SingularityWriter(parser.recipe)
        recipe = writer.convert()
        with open(singularityfile, "w") as sf:
            sf.write(recipe)
        return singularityfile

    @staticmethod
    def _get_recipe_file(build_ctx_path, cid):
        dockerfile = os.path.join(build_ctx_path, "Dockerfile")
        singularityfile = os.path.join(
            build_ctx_path, "Singularity.{}".format(cid[:-4])
        )

        if os.path.isfile(dockerfile):
            return SingularityRunner._convert(dockerfile, singularityfile)
        else:
            log.fail("No Dockerfile was found.")

    @staticmethod
    def _in_docker():
        """ Returns TRUE if we are being executed in a Docker container. """
        if os.path.isfile("/proc/1/cgroup"):
            with open("/proc/1/cgroup", "r") as f:
                return "docker" in f.read() or "lxc" in f.read()

    def _build_from_recipe(self, build_ctx_path, build_dest, cid):
        SingularityRunner.lock.acquire()
        pwd = os.getcwd()
        os.chdir(build_ctx_path)
        recipefile = SingularityRunner._get_recipe_file(build_ctx_path, cid)
        self._s.build(recipe=recipefile, image=cid, build_folder=build_dest, force=True)
        os.chdir(pwd)
        SingularityRunner.lock.release()

    def _setup_singularity_cache(self):
        self._singularity_cache = os.path.join(
            self._config.cache_dir, "singularity", self._config.wid
        )
        os.makedirs(self._singularity_cache, exist_ok=True)

    def _get_container_options(self):
        container_args = {
            "userns": True,
            "pwd": "/workspace",
            "bind": [f"{self._config.workspace_dir}:/workspace"],
        }

        self._update_with_engine_config(container_args)

        options = []
        for k, v in container_args.items():
            if isinstance(v, list):
                for item in v:
                    options.append(pu.key_value_to_flag(k, item))
            else:
                options.append(pu.key_value_to_flag(k, v))

        options = " ".join(options).split(" ")
        log.debug(f"container options: {options}\n")

        return options

    def _create_container(self, step, cid):
        build, image, _, _, build_ctx_path = self._get_build_info(step)

        if "shub://" in step.uses or "library://" in step.uses:
            build = False
            image = step.uses
            build_ctx_path = None

        if build:
            log.info(f"[{step.id}] singularity build {cid} {build_ctx_path}")
            if not self._config.dry_run:
                self._build_from_recipe(build_ctx_path, self._singularity_cache, cid)
        elif not self._config.skip_pull and not step.skip_pull:
            log.info(f"[{step.id}] singularity pull {cid} {image}")
            if not self._config.dry_run:
                self._s.pull(image=image, name=cid, pull_folder=self._singularity_cache)

    def _singularity_start(self, step, cid):
        env = self._prepare_environment(step)

        # set the environment variables
        for k, v in env.items():
            os.environ[k] = str(v)

        args = list(step.args)
        runs = list(step.runs)
        ecode = None

        if runs:
            info = f"[{step.id}] singularity exec {cid} {runs}"
            commands = runs
            start_fn = self._s.execute
        else:
            info = f"[{step.id}] singularity run {cid} {args}"
            commands = args
            start_fn = self._s.run

        log.info(info)

        if self._config.dry_run:
            return 0

        options = self._get_container_options()
        output = start_fn(self._container, commands, stream=True, options=options)
        try:
            for line in output:
                log.step_info(line.strip("\n"))
            ecode = 0
        except CalledProcessError as ex:
            ecode = ex.returncode

        return ecode

    def stop_running_tasks(self):
        pass



import os
import tempfile

import git

from popper import scm
from popper.cli import log

from .test_common import PopperTest


class TestScm(PopperTest):
    @classmethod
    def setUpClass(self):
        log.setLevel("CRITICAL")

    @classmethod
    def tearDownClass(self):
        log.setLevel("NOTSET")

    def test_empty_repo(self):
        tempdir = tempfile.mkdtemp()
        repo = git.Repo.init(tempdir)
        self.assertTrue(scm.is_empty(repo))

        self.assertIsNone(scm.get_sha(repo))
        self.assertIsNone(scm.get_branch(repo))
        self.assertIsNone(scm.get_tag(repo))

        repo = self.mk_repo()
        self.assertFalse(scm.is_empty(repo))

    def test_get_remote_url(self):
        repo = self.mk_repo()
        url = scm.get_remote_url(repo)
        self.assertEqual(url, "https://github.com/my/repo")

        self.assertEqual(scm.get_remote_url(None), "")

    def test_get_branch_in_detached_head_state(self):
        repo = self.mk_repo()

        repo.git.checkout("HEAD~1")

        os.environ["TRAVIS_BRANCH"] = "travis"
        self.assertEqual("travis", scm.get_branch(repo))
        os.environ.pop("TRAVIS_BRANCH")

        os.environ["GIT_BRANCH"] = "jenkins"
        self.assertEqual("jenkins", scm.get_branch(repo))
        os.environ.pop("GIT_BRANCH")

        os.environ["CIRCLE_BRANCH"] = "circle"
        self.assertEqual("circle", scm.get_branch(repo))
        os.environ.pop("CIRCLE_BRANCH")

        os.environ["CI_COMMIT_REF_NAME"] = "gitlab"
        self.assertEqual("gitlab", scm.get_branch(repo))
        os.environ.pop("CI_COMMIT_REF_NAME")

        self.assertEqual(scm.get_sha(repo), scm.get_branch(repo))

        # None given as arg
        self.assertIsNone(scm.get_sha(None))
        self.assertIsNone(scm.get_sha(None, short=8))
        self.assertIsNone(scm.get_branch(None))

    def test_get_tag(self):
        self.assertIsNone(scm.get_tag(None))

        repo = self.mk_repo()
        repo.git.checkout("HEAD~1")

        os.environ["TRAVIS_TAG"] = "travis"
        self.assertEqual("travis", scm.get_tag(repo))
        os.environ.pop("TRAVIS_TAG")

        os.environ["GIT_TAG"] = "jenkins"
        self.assertEqual("jenkins", scm.get_tag(repo))
        os.environ.pop("GIT_TAG")

        os.environ["CIRCLE_TAG"] = "circle"
        self.assertEqual("circle", scm.get_tag(repo))
        os.environ.pop("CIRCLE_TAG")

        os.environ["CI_COMMIT_REF_NAME"] = "gitlab"
        self.assertEqual("gitlab", scm.get_tag(repo))
        os.environ.pop("CI_COMMIT_REF_NAME")

        # without any of the above, it should be empty
        self.assertIsNone(scm.get_tag(repo))

        # test with a tagged commit
        repo = self.mk_repo(tag="foo")
        self.assertEqual("foo", scm.get_tag(repo))

    def test_clone(self):
        tempdir = tempfile.mkdtemp()
        tdir = os.path.join(tempdir, "test_clone")
        os.makedirs(tdir)
        currdir = os.getcwd()
        os.chdir(tdir)
        scm.clone(
            "https://github.com",
            "popperized",
            "github-actions-demo",
            os.path.join(os.getcwd(), "gad"),
            "develop",
        )
        repo = git.Repo(os.path.join(os.getcwd(), "gad"))
        self.assertEqual(repo.active_branch.name, "develop")
        scm.clone(
            "https://github.com",
            "popperized",
            "github-actions-demo",
            os.path.join(os.getcwd(), "gad"),
            "master",
        )
        repo = git.Repo(os.path.join(os.getcwd(), "gad"))
        self.assertEqual(repo.active_branch.name, "master")
        repo.close()
        os.chdir(currdir)

    def test_parse(self):
        test_url = "ssh://git@github.com:popperized" "/github-actions-demo.git"
        self.assertRaises(SystemExit, scm.parse, test_url)
        test_url = "https://github.com/popperized@master"
        self.assertRaises(SystemExit, scm.parse, test_url)
        test_url = "github.com"
        self.assertRaises(SystemExit, scm.parse, test_url)
        test_url = "@master"
        self.assertRaises(SystemExit, scm.parse, test_url)
        test_url = "http://gitlab.com/popperized/github-actions-demo.git"
        parts = scm.parse(test_url)
        self.assertTupleEqual(
            parts,
            (
                "http://gitlab.com",
                "gitlab.com",
                "popperized",
                "github-actions-demo",
                "",
                None,
            ),
        )
        test_url = "https://github.com/popperized" "/github-actions-demo@master"
        parts = scm.parse(test_url)
        self.assertTupleEqual(
            parts,
            (
                "https://github.com",
                "github.com",
                "popperized",
                "github-actions-demo",
                "",
                "master",
            ),
        )
        test_url = "github.com/popperized/github-actions-demo"
        parts = scm.parse(test_url)
        self.assertTupleEqual(
            parts,
            (
                "https://github.com",
                "github.com",
                "popperized",
                "github-actions-demo",
                "",
                None,
            ),
        )
        test_url = "popperized/github-actions-demo/path/to/action@develop"
        parts = scm.parse(test_url)
        self.assertTupleEqual(
            parts,
            (
                "https://github.com",
                "github.com",
                "popperized",
                "github-actions-demo",
                "path/to/action",
                "develop",
            ),
        )



import os
import shutil
import time
import unittest

import popper.scm as scm
import popper.utils as pu

from popper.config import ConfigLoader
from popper.runner import WorkflowRunner
from popper.runner_kubernetes import KubernetesRunner
from popper.cli import log as log
from .test_common import PopperTest

from kubernetes import config, client
from kubernetes.client import Configuration, V1DeleteOptions
from kubernetes.client.api import core_v1_api
from kubernetes.stream import stream

from box import Box


@unittest.skipIf(
    os.environ.get("ENABLE_K8S_RUNNER_TESTS", "0") != "1",
    "Kubernetes runner tests not enabled.",
)
class TestKubernetesRunner(PopperTest):
    def setUp(self):
        log.setLevel("CRITICAL")
        config.load_kube_config()

        c = Configuration()
        c.assert_hostname = False
        Configuration.set_default(c)
        self._kclient = core_v1_api.CoreV1Api()

        config.list_kube_config_contexts()

    def tearDown(self):
        log.setLevel("NOTSET")

    def test_vol_claim_create_delete(self):
        repo = self.mk_repo()
        conf = ConfigLoader.load(workspace_dir=repo.working_dir)
        with KubernetesRunner(config=conf) as kr:
            kr._vol_claim_create()
            response = self._kclient.read_namespaced_persistent_volume_claim(
                kr._vol_claim_name, namespace="default"
            )
            self.assertEqual(response.status.phase, "Bound")
            kr._vol_claim_delete()
            self.assertRaises(
                Exception,
                self._kclient.read_namespaced_persistent_volume_claim,
                {"name": kr._vol_claim_name, "namespace": "default"},
            )

        repo.close()
        shutil.rmtree(repo.working_dir, ignore_errors=True)

    def test_init_pod_create_delete(self):
        repo = self.mk_repo()
        conf = ConfigLoader.load(workspace_dir=repo.working_dir)
        with KubernetesRunner(config=conf) as kr:
            kr._vol_claim_create()
            kr._init_pod_create()
            time.sleep(5)
            response = self._kclient.read_namespaced_pod(
                kr._init_pod_name, namespace="default"
            )
            self.assertEqual(response.status.phase, "Running")
            kr._init_pod_delete()
            kr._vol_claim_delete()
            self.assertRaises(
                Exception,
                self._kclient.read_namespaced_pod,
                **{"name": kr._init_pod_name, "namespace": "default"},
            )

        repo.close()
        shutil.rmtree(repo.working_dir, ignore_errors=True)

    def test_copy_ctx(self):
        repo = self.mk_repo()
        conf = ConfigLoader.load(workspace_dir=repo.working_dir)
        with KubernetesRunner(config=conf) as kr:
            kr._vol_claim_create()
            kr._init_pod_create()
            time.sleep(5)
            response = self._kclient.read_namespaced_pod(
                kr._init_pod_name, namespace="default"
            )
            self.assertEqual(response.status.phase, "Running")
            kr._copy_ctx()
            kr._init_pod_delete()
            kr._vol_claim_delete()

        repo.close()
        shutil.rmtree(repo.working_dir, ignore_errors=True)

    def test_pod_create_delete_exitcode(self):
        repo = self.mk_repo()
        conf = ConfigLoader.load(workspace_dir=repo.working_dir)
        with KubernetesRunner(config=conf) as kr:
            kr._vol_claim_create()
            step = Box(
                {
                    "id": "test",
                    "uses": "docker://alpine:3.9",
                    "runs": ("echo", "hello"),
                },
                default_box=True,
            )
            kr._pod_name = kr._base_pod_name + f"-{step.id}"
            kr._pod_create(step, "alpine:3.9")
            self.assertEqual(kr._pod_exit_code(), 0)
            response = self._kclient.read_namespaced_pod(
                kr._pod_name, namespace="default"
            )
            self.assertEqual(response.status.phase, "Succeeded")
            kr._pod_delete()
            self.assertRaises(
                Exception,
                self._kclient.read_namespaced_pod,
                **{"name": kr._pod_name, "namespace": "default"},
            )

            time.sleep(5)

            step = Box(
                {
                    "id": "test",
                    "uses": "docker://alpine:3.9",
                    "runs": ("ecdho", "hello"),
                },
                default_box=True,
            )
            kr._pod_name = kr._base_pod_name + f"-{step.id}"
            kr._pod_create(step, "alpine:3.9")
            self.assertEqual(kr._pod_exit_code(), 1)
            response = self._kclient.read_namespaced_pod(
                kr._pod_name, namespace="default"
            )
            self.assertEqual(response.status.phase, "Failed")
            kr._pod_delete()
            kr._vol_claim_delete()

            self.assertRaises(
                Exception,
                self._kclient.read_namespaced_pod,
                **{"name": kr._pod_name, "namespace": "default"},
            )



import unittest

from popper import utils as pu
from popper.cli import log


class TestUtils(unittest.TestCase):
    def setUp(self):
        log.setLevel("CRITICAL")

    def tearDown(self):
        log.setLevel("NOTSET")

    def test_sanitized_name(self):
        name = "test action"
        santizied_name = pu.sanitized_name(name, "1234")
        self.assertEqual(santizied_name, "popper_test_action_1234")

        name = "test@action"
        santizied_name = pu.sanitized_name(name, "1234")
        self.assertEqual(santizied_name, "popper_test_action_1234")

        name = "test(action)"
        santizied_name = pu.sanitized_name(name, "1234")
        self.assertEqual(santizied_name, "popper_test_action__1234")

    def test_assert_executable_exists(self):
        pu.assert_executable_exists("ls")
        self.assertRaises(SystemExit, pu.assert_executable_exists, "abcd")

    def test_kv_to_flag(self):
        self.assertEqual(pu.key_value_to_flag("x", "a"), "-x a")
        self.assertEqual(pu.key_value_to_flag("y", True), "-y")
        self.assertEqual(pu.key_value_to_flag("y", False), "")
        self.assertEqual(pu.key_value_to_flag("yy", True), "--yy")
        self.assertEqual(pu.key_value_to_flag("zz", "c"), "--zz c")
        eq = True
        self.assertEqual(pu.key_value_to_flag("x", "a", eq), "-x=a")
        self.assertEqual(pu.key_value_to_flag("y", True, eq), "-y=true")
        self.assertEqual(pu.key_value_to_flag("y", False, eq), "-y=false")
        self.assertEqual(pu.key_value_to_flag("yy", True, eq), "--yy=true")
        self.assertEqual(pu.key_value_to_flag("zz", "c", eq), "--zz=c")



import click
import traceback

from popper.cli import pass_context, log
from popper.config import ConfigLoader
from popper.parser import WorkflowParser
from popper.runner import WorkflowRunner


@click.command("sh", short_help="Start an interactive shell for the given step.")
@click.argument("step", required=True)
@click.option(
    "-f",
    "--file",
    help="File containing the definition of the workflow.",
    required=False,
    default=".popper.yml",
)
@click.option(
    "-e",
    "--entrypoint",
    help="Shell program to invoke.",
    required=False,
    default="/bin/bash",
)
@pass_context
def cli(ctx, file, step, entrypoint):
    """Opens an interactive shell using all the attributes defined in the workflow file
    for the given STEP, but ignoring ``runs`` and ``args``. By default, it invokes
    /bin/bash. If you need to invoke another one, you can specify it in the --entrypoint
    flag.

    NOTE: this command only works for (local) host runner in Docker.
    """
    wf = WorkflowParser.parse(file=file, step=step, immutable=False)

    # override entrypoint
    step = wf.steps[0]
    step.args = []
    step.runs = entrypoint

    # configure runner so containers execute in attached mode and create a tty
    config = ConfigLoader.load(engine_name="docker", pty=True)

    with WorkflowRunner(config) as runner:
        try:
            runner.run(wf)
        except Exception as e:
            log.debug(traceback.format_exc())
            log.fail(e)



import os
import time
import signal
import socket
import threading

from popper import utils as pu
from popper.cli import log as log
from popper.runner_host import HostRunner
from popper.runner_host import SingularityRunner as HostSingularityRunner


class SlurmRunner(HostRunner):
    def __init__(self, **kw):
        super(SlurmRunner, self).__init__(**kw)
        self._spawned_jobs = set()

    def __exit__(self, exc_type, exc, traceback):
        self._spawned_jobs = set()

    def _tail_output(self, out_file):
        self._out_stream_pid = set()
        _, ecode, _ = HostRunner._exec_cmd(
            ["tail", "-f", out_file], pids=self._out_stream_pid
        )
        return ecode

    def _start_out_stream(self, out_file):
        self._out_stream_thread = threading.Thread(
            target=self._tail_output, args=(out_file,)
        )
        self._out_stream_thread.start()
        # give time so that _exec_cmd puts the pid inside the self._out_stream_pid set
        time.sleep(1)

    def _stop_out_stream(self):
        if len(self._out_stream_pid) != 1:
            log.fail("Cannot find PID for tail process")
        _out_stream_pid = list(self._out_stream_pid)[0]
        try:
            os.kill(_out_stream_pid, signal.SIGKILL)
        except ProcessLookupError:
            log.warning("Tail process was stopped by some other process.")
        self._out_stream_thread.join()

    def _set_config_vars(self, step):
        self._nodes = self._config.resman_opts.get(step.id, {}).get("nodes", 1)
        self._nodelist = self._config.resman_opts.get(step.id, {}).get("nodelist", None)
        self._ntasks = self._config.resman_opts.get(step.id, {}).get(
            "ntasks", self._nodes
        )
        self._ntasks_per_node = self._config.resman_opts.get(step.id, {}).get(
            "ntasks-per-node", 1
        )

    def _get_resman_kwargs(self, step):
        default_options = ["nodes", "nodelist", "ntasks", "ntasks-per-node"]
        resman_options = []
        for k, v in self._config.resman_opts.get(step.id, {}).items():
            if k not in default_options:
                flag = pu.key_value_to_flag(k, v)
                if flag:
                    resman_options.extend(flag.split())

        return resman_options

    def _exec_srun(self, cmd, step, **kwargs):
        self._set_config_vars(step)
        _cmd = [
            "srun",
            "--nodes",
            f"{self._nodes}",
            "--ntasks",
            f"{self._ntasks}",
            "--ntasks-per-node",
            f"{self._ntasks_per_node}",
        ]

        if self._nodelist:
            _cmd.extend(["--nodelist", self._nodelist])

        _cmd.extend(self._get_resman_kwargs(step))
        _cmd.extend(cmd)

        log.debug(f"Command: {_cmd}")

        if self._config.dry_run:
            return 0

        _, ecode, _ = HostRunner._exec_cmd(_cmd, **kwargs)
        return ecode

    def _exec_mpi(self, cmd, step, **kwargs):
        self._set_config_vars(step)
        job_name = pu.sanitized_name(step.id, self._config.wid)
        mpi_cmd = ["mpirun", f"{' '.join(cmd)}"]

        job_script = os.path.join(f"{job_name}.sh")
        out_file = os.path.join(f"{job_name}.out")

        with open(out_file, "w"):
            pass

        with open(job_script, "w") as f:
            f.write("#!/bin/bash\n")
            f.write(f"#SBATCH --job-name={job_name}\n")
            f.write(f"#SBATCH --output={out_file}\n")
            f.write(f"#SBATCH --nodes={self._nodes}\n")
            f.write(f"#SBATCH --ntasks={self._ntasks}\n")
            f.write(f"#SBATCH --ntasks-per-node={self._ntasks_per_node}\n")
            if self._nodelist:
                f.write(f"#SBATCH --nodelist={self._nodelist}\n")
            f.write(" ".join(mpi_cmd))

        sbatch_cmd = [
            "sbatch",
            "--wait",
        ]
        sbatch_cmd.extend(self._get_resman_kwargs(step))
        sbatch_cmd.extend([job_script])

        log.debug(f"Command: {sbatch_cmd}")

        if self._config.dry_run:
            return 0

        self._spawned_jobs.add(job_name)
        self._start_out_stream(out_file)

        _, ecode, _ = HostRunner._exec_cmd(sbatch_cmd, **kwargs)

        self._stop_out_stream()
        self._spawned_jobs.remove(job_name)

        return ecode

    def stop_running_tasks(self):
        for job_name in self._spawned_jobs:
            log.info(f"Cancelling job {job_name}")
            _, ecode, _ = HostRunner._exec_cmd(["scancel", "--name", job_name])
            if ecode != 0:
                log.warning(f"Failed to cancel the job {job_name}.")


class SingularityRunner(SlurmRunner, HostSingularityRunner):
    def __init__(self, **kw):
        super(SingularityRunner, self).__init__(init_spython_client=False, **kw)
        if self._config.reuse:
            log.fail("Reuse not supported for SingularityRunner.")

    def __exit__(self, exc_type, exc_value, exc_traceback):
        pass

    def run(self, step):
        self._setup_singularity_cache()
        cid = pu.sanitized_name(step.id, self._config.wid) + ".sif"
        self._container = os.path.join(self._singularity_cache, cid)

        build, img, _, _, build_ctx_path = self._get_build_info(step)

        if "shub://" in step.uses or "library://" in step.uses:
            build = False
            img = step.uses
            build_ctx_path = None

        self._exec_srun(["rm", "-rf", self._container], step)

        if build:
            recipefile = self._get_recipe_file(build_ctx_path, cid)
            log.info(f"[{step.id}] srun singularity build {self._container}")
            self._exec_srun(
                ["singularity", "build", "--fakeroot", self._container, recipefile,],
                step,
                cwd=os.path.dirname(recipefile),
            )
        else:
            log.info(f"[{step.id}] srun singularity pull {self._container}")
            self._exec_srun(["singularity", "pull", self._container, img], step)

        cmd = self._create_cmd(step, cid)
        self._spawned_containers.add(cid)

        if self._config.resman_opts.get(step.id, {}).get("mpi", True):
            log.info(f'[{step.id}] sbatch {" ".join(cmd)}')
            ecode = self._exec_mpi(cmd, step)
        else:
            log.info(f'[{step.id}] srun {" ".join(cmd)}')
            ecode = self._exec_srun(cmd, step, logging=True)

        self._spawned_containers.remove(cid)
        return ecode

    def _create_cmd(self, step, cid):
        env = self._prepare_environment(step)
        for k, v in env.items():
            os.environ[k] = str(v)

        if step.runs:
            commands = step.runs
            cmd = ["singularity", "exec"]
        else:
            commands = step.args
            cmd = ["singularity", "run"]

        options = self._get_container_options()

        cmd.extend(options)
        cmd.extend([self._container])
        cmd.extend(commands)

        return cmd



import os
import shutil
import time
import unittest

from testfixtures import LogCapture
from subprocess import Popen

import popper.utils as pu

from popper.config import ConfigLoader
from popper.parser import WorkflowParser
from popper.runner import WorkflowRunner
from popper.runner_host import HostRunner, DockerRunner, PodmanRunner, SingularityRunner
from popper.cli import log as log
from .test_common import PopperTest

import docker

from box import Box


class TestHostHostRunner(PopperTest):
    def setUp(self):
        log.setLevel("CRITICAL")

    def test_host_run(self):

        repo = self.mk_repo()
        conf = ConfigLoader.load(workspace_dir=repo.working_dir)

        with WorkflowRunner(conf) as r:
            wf_data = {
                "steps": [{"uses": "sh", "runs": ["cat"], "args": ["README.md"],}]
            }
            r.run(WorkflowParser.parse(wf_data=wf_data))

            wf_data = {
                "steps": [
                    {
                        "uses": "sh",
                        "runs": ["bash", "-c", "echo $FOO > hello.txt ; pwd"],
                        "env": {"FOO": "bar"},
                    }
                ]
            }
            r.run(WorkflowParser.parse(wf_data=wf_data))
            with open(os.path.join(repo.working_dir, "hello.txt"), "r") as f:
                self.assertEqual(f.read(), "bar\n")

            wf_data = {"steps": [{"uses": "sh", "runs": ["nocommandisnamedlikethis"]}]}
            self.assertRaises(SystemExit, r.run, WorkflowParser.parse(wf_data=wf_data))

            # check exit code 78
            wf_data = {
                "steps": [
                    {"uses": "sh", "runs": ["touch", "one.txt"]},
                    {"uses": "sh", "runs": ["bash", "-c", "exit 78"]},
                    {"uses": "sh", "runs": ["touch", "two.txt"]},
                ]
            }
            r.run(WorkflowParser.parse(wf_data=wf_data))
            self.assertTrue(os.path.isfile(os.path.join(repo.working_dir, "one.txt")))
            self.assertFalse(os.path.isfile(os.path.join(repo.working_dir, "two.txt")))

        repo.close()
        shutil.rmtree(repo.working_dir, ignore_errors=True)

    def test_exec_cmd(self):
        cmd = ["echo", "hello-world"]
        pid, ecode, output = HostRunner._exec_cmd(cmd, logging=False)
        self.assertGreater(pid, 0)
        self.assertEqual(ecode, 0)
        self.assertEqual(output, "hello-world")

        with LogCapture("popper") as logc:
            pid, ecode, output = HostRunner._exec_cmd(cmd)
            self.assertGreater(pid, 0)
            self.assertEqual(ecode, 0)
            self.assertEqual(output, "")
            logc.check_present(("popper", "STEP_INFO", "hello-world"))

        cmd = ["env"]
        pid, ecode, output = HostRunner._exec_cmd(
            cmd, env={"TESTACION": "test"}, cwd="/tmp", logging=False
        )
        self.assertGreater(pid, 0)
        self.assertEqual(ecode, 0)
        self.assertTrue("TESTACION" in output)

        _pids = set()
        _, _, _ = HostRunner._exec_cmd(["sleep", "2"], pids=_pids)
        self.assertEqual(len(_pids), 1)

    def test_stop_running_tasks(self):
        with HostRunner() as hr:
            with Popen(["sleep", "2000"]) as p:
                pid = p.pid
                hr._spawned_pids.add(pid)
                hr.stop_running_tasks()
        time.sleep(2)
        self.assertRaises(ProcessLookupError, os.kill, pid, 0)


@unittest.skipIf(os.environ.get("ENGINE", "docker") != "docker", "ENGINE != docker")
class TestHostDockerRunner(PopperTest):
    def setUp(self):
        log.setLevel("CRITICAL")

    def test_create_container(self):
        config = ConfigLoader.load()
        step = Box(
            {
                "uses": "docker://alpine:3.9",
                "runs": ["echo hello"],
                "id": "kontainer_one",
            },
            default_box=True,
        )
        cid = pu.sanitized_name(step.id, config.wid)
        with DockerRunner(init_docker_client=True, config=config) as dr:
            c = dr._create_container(cid, step)
            self.assertEqual(c.status, "created")
            c.remove()

        step = Box(
            {
                "uses": "docker://alpine:3.9",
                "runs": ["echo", "hello_world"],
                "id": "KoNtAiNeR tWo",
            },
            default_box=True,
        )
        cid = pu.sanitized_name(step.id, config.wid)
        with DockerRunner(init_docker_client=True, config=config) as dr:
            c = dr._create_container(cid, step)
            self.assertEqual(c.status, "created")
            c.remove()

        step = Box(
            {
                "uses": "docker://alpine:3.9",
                "runs": ["echo", "hello_world"],
                "options": {"volumes": ("/tmp:/bindmounteddir",)},
                "id": "container",
            },
            default_box=True,
        )
        cid = pu.sanitized_name(step.id, config.wid)
        with DockerRunner(init_docker_client=True, config=config) as dr:
            c = dr._create_container(cid, step)
            self.assertEqual(c.status, "created")
            c.remove()

    def test_stop_running_tasks(self):
        with DockerRunner() as dr:
            dclient = docker.from_env()
            c1 = dclient.containers.run(
                "debian:buster-slim", "sleep 20000", detach=True
            )
            c2 = dclient.containers.run("alpine:3.9", "sleep 10000", detach=True)
            dr._spawned_containers.add(c1)
            dr._spawned_containers.add(c2)
            dr.stop_running_tasks()
            self.assertEqual(c1.status, "created")
            self.assertEqual(c2.status, "created")
            dclient.close()

    def test_docker_basic_run(self):
        repo = self.mk_repo()
        conf = ConfigLoader.load(workspace_dir=repo.working_dir)
        test_string = "STEP_INFO:popper:Successfully tagged popperized/bin:master"

        with WorkflowRunner(conf) as r:
            wf_data = {"steps": [{"uses": "popperized/bin/sh@master", "args": ["ls"],}]}
            with self.assertLogs(log, level="STEP_INFO") as cm:
                r.run(WorkflowParser.parse(wf_data=wf_data))
            self.assertTrue(test_string in cm.output, f"Got cmd output: {cm.output}")

            wf_data = {
                "steps": [
                    {
                        "uses": "docker://alpine:3.9",
                        "args": ["sh", "-c", "echo $FOO > hello.txt ; pwd"],
                        "env": {"FOO": "bar"},
                    }
                ]
            }
            r.run(WorkflowParser.parse(wf_data=wf_data))
            with open(os.path.join(repo.working_dir, "hello.txt"), "r") as f:
                self.assertEqual(f.read(), "bar\n")

            wf_data = {
                "steps": [
                    {
                        "uses": "docker://alpine:3.9",
                        "args": ["nocommandisnamedlikethis"],
                    }
                ]
            }
            self.assertRaises(SystemExit, r.run, WorkflowParser.parse(wf_data=wf_data))

        conf = ConfigLoader.load(workspace_dir=repo.working_dir, quiet=True)
        with WorkflowRunner(conf) as r:
            wf_data = {"steps": [{"uses": "popperized/bin/sh@master", "args": ["ls"],}]}
            with self.assertLogs(log, level="STEP_INFO") as cm:
                r.run(WorkflowParser.parse(wf_data=wf_data))
            self.assertTrue(test_string not in cm.output)

        repo.close()
        shutil.rmtree(repo.working_dir, ignore_errors=True)


@unittest.skipIf(os.environ.get("ENGINE", "docker") != "podman", "ENGINE != podman")
class TestHostPodmanRunner(PopperTest):
    def setUp(self):
        log.setLevel("CRITICAL")

    def test_stop_running_tasks(self):
        with PodmanRunner() as pr:
            cmd = ["podman", "run", "-d", "-q"]
            _, _, c1 = HostRunner._exec_cmd(
                cmd + ["debian:buster-slim", "sleep", "20000"], logging=False,
            )
            _, _, c2 = HostRunner._exec_cmd(
                cmd + ["alpine:3.9", "sleep", "10000"], logging=False
            )
            c1 = c1.strip()
            c2 = c2.strip()
            pr._spawned_containers.add(c1)
            pr._spawned_containers.add(c2)
            pr.stop_running_tasks()
            status_cmd = [
                "podman",
                "container",
                "inspect",
                "-f",
                str("{{.State.Status}}"),
            ]
            c1_status_cmd = status_cmd + [c1]
            c2_status_cmd = status_cmd + [c2]
            _, _, c1_status = HostRunner._exec_cmd(c1_status_cmd, logging=False)
            _, _, c2_status = HostRunner._exec_cmd(c2_status_cmd, logging=False)
            self.assertEqual(c1_status, "exited")
            self.assertEqual(c2_status, "exited")

    def test_find_container(self):
        config = ConfigLoader.load()
        step = Box(
            {
                "uses": "docker://alpine:3.9",
                "runs": ["echo hello"],
                "id": "kontainer_one",
            },
            default_box=True,
        )
        cid = pu.sanitized_name(step.id, config.wid)
        with PodmanRunner(init_podman_client=True, config=config) as pr:
            c = pr._find_container(cid)
            self.assertEqual(c, None)
        with PodmanRunner(init_podman_client=True, config=config) as pr:
            container = pr._create_container(cid, step)
            c = pr._find_container(cid)
            self.assertEqual(c, container)
            cmd = ["podman", "container", "rm", "-f", cid]
            HostRunner._exec_cmd(cmd, logging=False)

    def test_create_container(self):
        config = ConfigLoader.load()
        step = Box(
            {
                "uses": "docker://alpine:3.9",
                "runs": ["echo hello"],
                "id": "kontainer_one",
            },
            default_box=True,
        )
        cid = pu.sanitized_name(step.id, config.wid)
        with PodmanRunner(init_podman_client=True, config=config) as pr:
            c = pr._create_container(cid, step)
            c_status_cmd = [
                "podman",
                "container",
                "inspect",
                "-f",
                str("{{.State.Status}}"),
                c,
            ]
            __, _, c_status = HostRunner._exec_cmd(c_status_cmd, logging=False)
            self.assertEqual(c_status, "configured")
            cmd = ["podman", "container", "rm", c]
            HostRunner._exec_cmd(cmd, logging=False)
        step = Box(
            {
                "uses": "docker://alpine:3.9",
                "runs": ["echo", "hello_world"],
                "id": "KoNtAiNeR tWo",
            },
            default_box=True,
        )
        cid = pu.sanitized_name(step.id, config.wid)
        with PodmanRunner(init_podman_client=True, config=config) as pr:
            c = pr._create_container(cid, step)
            c_status_cmd = [
                "podman",
                "container",
                "inspect",
                "-f",
                str("{{.State.Status}}"),
                c,
            ]
            __, _, c_status = HostRunner._exec_cmd(c_status_cmd, logging=False)
            self.assertEqual(c_status, "configured")
            cmd = ["podman", "container", "rm", c]
            HostRunner._exec_cmd(cmd, logging=False)

    def test_podman_basic_run(self):
        repo = self.mk_repo()
        conf = ConfigLoader.load(engine_name="podman", workspace_dir=repo.working_dir)

        with WorkflowRunner(conf) as r:
            wf_data = {"steps": [{"uses": "popperized/bin/sh@master", "args": ["ls"],}]}
            r.run(WorkflowParser.parse(wf_data=wf_data))

            wf_data = {
                "steps": [
                    {
                        "uses": "docker://alpine:3.9",
                        "args": ["sh", "-c", "echo $FOO > hello.txt ; pwd"],
                        "env": {"FOO": "bar"},
                    }
                ]
            }
            r.run(WorkflowParser.parse(wf_data=wf_data))
            with open(os.path.join(repo.working_dir, "hello.txt"), "r") as f:
                self.assertEqual(f.read(), "bar\n")

            wf_data = {
                "steps": [
                    {
                        "uses": "docker://alpine:3.9",
                        "args": ["nocommandisnamedlikethis"],
                    }
                ]
            }
            self.assertRaises(SystemExit, r.run, WorkflowParser.parse(wf_data=wf_data))

            # wf_data = {
            #     "steps": [
            #         {
            #             "uses": "docker://golang:1.14.4-alpine3.11",
            #             "args": ["go", "build", "-v", "."],
            #         }
            #     ]
            # }
            # self.assertRaises(SystemExit, r.run, WorkflowParser.parse(wf_data=wf_data))

        repo.close()
        shutil.rmtree(repo.working_dir, ignore_errors=True)


@unittest.skipIf(
    os.environ.get("ENGINE", "docker") != "singularity", "ENGINE != singularity"
)
class TestHostSingularityRunner(PopperTest):
    def setUp(self):
        log.setLevel("CRITICAL")

    def test_get_recipe_file(self):
        repo = self.mk_repo()
        build_ctx_path = repo.working_dir

        with open(os.path.join(build_ctx_path, "Dockerfile"), "w") as f:
            f.write(
                """
FROM alpine
RUN apk update && apk add bash
ADD README.md /
ENTRYPOINT ["/bin/bash"]"""
            )

        singularity_file = SingularityRunner._get_recipe_file(
            build_ctx_path, "sample.sif"
        )
        self.assertEqual(
            singularity_file, os.path.join(build_ctx_path, "Singularity.sample")
        )
        self.assertEqual(os.path.exists(singularity_file), True)
        with open(singularity_file) as f:
            self.assertEqual(
                f.read(),
                '''Bootstrap: docker
From: alpine
%files
README.md /
%post

apk update && apk add bash
%runscript
exec /bin/bash "$@"
%startscript
exec /bin/bash "$@"''',
            )

        os.remove(os.path.join(build_ctx_path, "Dockerfile"))
        self.assertRaises(
            SystemExit, SingularityRunner._get_recipe_file, build_ctx_path, "sample.sif"
        )

    def test_create_container(self):
        config = ConfigLoader.load()
        step_one = Box(
            {
                "uses": "docker://alpine:3.9",
                "runs": ["echo hello"],
                "id": "kontainer_one",
            },
            default_box=True,
        )

        step_two = Box(
            {
                "uses": "popperized/bin/sh@master",
                "args": ["ls"],
                "id": "kontainer_two",
            },
            default_box=True,
        )

        cid_one = pu.sanitized_name(step_one.id, config.wid)
        cid_two = pu.sanitized_name(step_two.id, config.wid)

        with SingularityRunner(config=config) as sr:
            sr._setup_singularity_cache()
            sr._create_container(step_one, cid_one)
            self.assertEqual(
                os.path.exists(os.path.join(sr._singularity_cache, cid_one)), True
            )
            os.remove(os.path.join(sr._singularity_cache, cid_one))

        with SingularityRunner(config=config) as sr:
            sr._setup_singularity_cache()
            sr._create_container(step_one, cid_two)
            self.assertEqual(
                os.path.exists(os.path.join(sr._singularity_cache, cid_two)), True
            )
            os.remove(os.path.join(sr._singularity_cache, cid_two))

    def test_setup_singularity_cache(self):
        config = ConfigLoader.load()
        with SingularityRunner(config=config) as sr:
            sr._setup_singularity_cache()
            self.assertEqual(
                f'{os.environ["HOME"]}/.cache/popper/singularity/{config.wid}',
                sr._singularity_cache,
            )

    def test_get_container_options(self):
        config_dict = {
            "engine": {
                "name": "singularity",
                "options": {
                    "hostname": "popper.local",
                    "ipc": True,
                    "bind": ["/path/in/host:/path/in/container"],
                },
            }
        }

        config = ConfigLoader.load(config_file=config_dict)
        with SingularityRunner(config=config) as sr:
            sr._setup_singularity_cache()
            options = sr._get_container_options()
            self.assertEqual(
                options,
                [
                    "--userns",
                    "--pwd",
                    "/workspace",
                    "--bind",
                    f"{os.getcwd()}:/workspace",
                    "--bind",
                    "/path/in/host:/path/in/container",
                    "--hostname",
                    "popper.local",
                    "--ipc",
                ],
            )

    def test_singularity_start(self):
        repo = self.mk_repo()
        conf = ConfigLoader.load(
            engine_name="singularity", workspace_dir=repo.working_dir
        )

        step = Box(
            {
                "uses": "docker://alpine:3.9",
                "runs": ["echo", "hello"],
                "name": "test_1",
            },
            default_box=True,
        )
        cid = pu.sanitized_name(step["name"], conf.wid)
        with SingularityRunner(config=conf) as sr:
            sr._setup_singularity_cache()
            sr._container = os.path.join(sr._singularity_cache, cid)
            sr._create_container(step, cid)
            self.assertEqual(sr._singularity_start(step, cid), 0)

        # step = Box({
        #     'uses': 'library://library/default/alpine:3.7',
        #     'runs': ['echo', 'hello'],
        #     'name': 'test_2'
        # }, default_box=True)
        # cid = pu.sanitized_name(step['name'], conf.wid)
        # with SingularityRunner(config=conf) as sr:
        #     sr._setup_singularity_cache()
        #     sr._container = os.path.join(sr._singularity_cache, cid)
        #     sr._create_container(step, cid)
        #     self.assertEqual(sr._singularity_start(step, cid), 0)

        # step = Box({
        #     'uses': 'shub://divetea/debian:latest',
        #     'runs': ['echo', 'hello'],
        #     'name': 'test_3'
        # }, default_box=True)
        # cid = pu.sanitized_name(step['name'], conf.wid)
        # with SingularityRunner(config=conf) as sr:
        #     sr._setup_singularity_cache()
        #     sr._container = os.path.join(sr._singularity_cache, cid)
        #     sr._create_container(step, cid)
        #     self.assertEqual(sr._singularity_start(step, cid), 0)

        step = Box(
            {
                "uses": "docker://alpine:3.9",
                "runs": ["ecdhoo", "hello"],
                "name": "test_4",
            },
            default_box=True,
        )
        cid = pu.sanitized_name(step["name"], conf.wid)
        with SingularityRunner(config=conf) as sr:
            sr._setup_singularity_cache()
            sr._container = os.path.join(sr._singularity_cache, cid)
            sr._create_container(step, cid)
            self.assertNotEqual(sr._singularity_start(step, cid), 0)

        with WorkflowRunner(conf) as r:
            wf_data = {"steps": [{"uses": "popperized/bin/sh@master", "args": ["ls"],}]}
            r.run(WorkflowParser.parse(wf_data=wf_data))

            wf_data = {
                "steps": [
                    {
                        "uses": "docker://alpine:3.9",
                        "args": ["sh", "-c", "echo $FOO > hello.txt ; pwd"],
                        "env": {"FOO": "bar"},
                    }
                ]
            }
            r.run(WorkflowParser.parse(wf_data=wf_data))
            with open(os.path.join(repo.working_dir, "hello.txt"), "r") as f:
                self.assertEqual(f.read(), "bar\n")

            wf_data = {
                "steps": [
                    {
                        "uses": "docker://alpine:3.9",
                        "args": ["nocommandisnamedlikethis"],
                    }
                ]
            }
            self.assertRaises(SystemExit, r.run, WorkflowParser.parse(wf_data=wf_data))

        repo.close()



import unittest

from popper.parser import WorkflowParser
from popper.cli import log


class TestWorkflow(unittest.TestCase):
    def setUp(self):
        log.setLevel("CRITICAL")

    def tearDown(self):
        log.setLevel("NOTSET")

    def test_empty_file(self):
        try:
            f = open("test.yml", "a")
            f.close()
            WorkflowParser.parse(file="test.yml")
            self.assertTrue(False, "Empty Workflow file does not raise an exception.")
        except SystemExit:
            self.assertTrue(True)
        else:
            self.assertTrue(False, "Empty Workflow file does not raise a SystemExit.")

    def test_new_workflow(self):
        wf_data = {}
        self.assertRaises(SystemExit, WorkflowParser.parse, **{"wf_data": wf_data})

        wf_data = {"unexpected": []}
        self.assertRaises(SystemExit, WorkflowParser.parse, **{"wf_data": wf_data})

        wf_data = {
            "steps": [
                {
                    "uses": "foo",
                    "id": "step",
                    "env": {"EN": "EE"},
                    "secrets": ["S"],
                    "dir": "/path/to/",
                    "options": {"name": "spam"},
                },
                {"uses": "bar", "runs": ["a", "b"], "args": ["c"], "skip_pull": True},
            ],
            "options": {"env": {"FOO": "bar"}, "secrets": ["Z"],},
        }
        wf = WorkflowParser.parse(wf_data=wf_data)

        step = wf.steps[0]
        self.assertEqual("step", step.id)
        self.assertEqual("foo", step.uses)
        self.assertEqual(("Z", "S"), step.secrets)
        self.assertEqual({"EN": "EE", "FOO": "bar"}, step.env)
        self.assertEqual("/path/to/", step.dir)
        self.assertEqual("spam", step.options.name)
        self.assertTrue(not step.runs)
        self.assertTrue(not step.args)
        self.assertFalse(step.skip_pull)

        step = wf.steps[1]
        self.assertEqual("bar", step.uses)
        self.assertEqual(("a", "b"), step.runs)
        self.assertEqual(("c",), step.args)
        self.assertTrue(step.skip_pull)
        self.assertTrue(not step.dir)
        self.assertEqual({"FOO": "bar"}, step.env)
        self.assertEqual(("Z",), step.secrets)
        self.assertEqual({"FOO": "bar"}, wf.options.env)
        self.assertEqual(("Z",), wf.options.secrets)

    def test_filter_all_but_given_step(self):
        wf_data = {
            "steps": [
                {"uses": "foo", "id": "one"},
                {"uses": "bar", "id": "two"},
                {"uses": "baz", "id": "three"},
            ]
        }
        wf = WorkflowParser.parse(wf_data=wf_data, step="two")
        self.assertEqual(1, len(wf.steps))
        self.assertEqual("two", wf.steps[0].id)
        self.assertEqual("bar", wf.steps[0].uses)

        # non-existing name
        self.assertRaises(
            SystemExit, WorkflowParser.parse, **{"wf_data": wf_data, "step": "four"}
        )

        # without id
        wf_data = {"steps": [{"uses": "foo"}, {"uses": "bar"}, {"uses": "baz"},]}
        wf = WorkflowParser.parse(wf_data=wf_data, step="2")
        self.assertEqual(1, len(wf.steps))
        self.assertEqual("2", wf.steps[0].id)

    def test_skip_steps(self):
        wf_data = {
            "steps": [
                {"uses": "foo", "id": "one"},
                {"uses": "bar", "id": "two"},
                {"uses": "baz", "id": "three"},
            ]
        }
        # skip one step
        wf = WorkflowParser.parse(wf_data=wf_data, skipped_steps=["two"])
        self.assertEqual(2, len(wf.steps))
        self.assertEqual("one", wf.steps[0].id)
        self.assertEqual("three", wf.steps[1].id)

        # more than one
        wf = WorkflowParser.parse(wf_data=wf_data, skipped_steps=["one", "three"])
        self.assertEqual(1, len(wf.steps))
        self.assertEqual("two", wf.steps[0].id)

        # non-existing name
        self.assertRaises(
            SystemExit,
            WorkflowParser.parse,
            **{"wf_data": wf_data, "skipped_steps": ["four"]}
        )

        # skip one step
        wf = WorkflowParser.parse(wf_data=wf_data, skipped_steps=["two"])
        self.assertEqual(2, len(wf.steps))
        self.assertEqual("one", wf.steps[0].id)
        self.assertEqual("three", wf.steps[1].id)

        # without id
        wf_data = {"steps": [{"uses": "foo"}, {"uses": "bar"}, {"uses": "baz"},]}
        wf = WorkflowParser.parse(wf_data=wf_data, skipped_steps=["1", "3"])
        self.assertEqual(1, len(wf.steps))
        self.assertEqual("2", wf.steps[0].id)

    def test_add_missing_ids(self):
        wf_data = {"steps": [{"uses": "foo"}, {"uses": "bar"}]}
        # skip one step
        wf = WorkflowParser.parse(wf_data=wf_data)
        self.assertEqual("1", wf.steps[0].id)
        self.assertEqual("2", wf.steps[1].id)

    def test_substitutions(self):
        # test wrong format for substitution key
        wf_data = {"steps": [{"uses": "whatever"}]}
        self.assertRaises(
            SystemExit,
            WorkflowParser.parse,
            **{"wf_data": wf_data, "substitutions": ["SUB1=WRONG"]}
        )

        # expect error when not all given subs are used
        wf_data = {
            "steps": [
                {
                    "uses": "some_$_SUB1",
                    "id": "some other $_SUB2",
                    "env": {"FOO": "env_$_SUB3"},
                    "secrets": ["secret_$_SUB4"],
                }
            ]
        }
        substitutions = [
            "_SUB1=ONE",
            "_SUB2=TWO",
            "_SUB3=THREE",
            "_SUB4=4",
            "_SUB5=UNUSED",
        ]
        self.assertRaises(
            SystemExit,
            WorkflowParser.parse,
            **{"wf_data": wf_data, "substitutions": substitutions}
        )

        # allow loose substitutions
        wf = WorkflowParser.parse(
            wf_data=wf_data, substitutions=substitutions, allow_loose=True
        )
        step = wf.steps[0]
        self.assertEqual("some_ONE", step.uses)
        self.assertEqual("some other TWO", step.id)
        self.assertEqual("env_THREE", step.env["FOO"])
        self.assertEqual(("secret_4",), step.secrets)



import click
import os

from popper.cli import pass_context, log

ci_files = {
    # ##########################################################3
    "travis": {
        "./.travis.yml": """
---
dist: xenial
language: python
python: 3.7
services: docker
install:
- git clone https://github.com/systemslab/popper /tmp/popper
- export PYTHONUNBUFFERED=1
- pip install /tmp/popper/cli
script: popper run -f {}
"""
    },
    # ##########################################################3
    "circle": {
        "./.circleci/config.yml": """
---
version: 2
jobs:
  build:
    machine: true
    steps:
    - checkout
    - run:
        command: |
        git clone https://github.com/systemslab/popper /tmp/popper
        export PYTHONUNBUFFERED=1
        pip install /tmp/popper/cli
        popper run -f {}
"""
    },
    # ##########################################################3
    "jenkins": {
        "./Jenkinsfile": """
---
stage ('Popper') {{ node {{
  sh "git clone https://github.com/systemslab/popper /tmp/popper"
  sh "export PYTHONUNBUFFERED=1"
  sh "pip install /tmp/popper/cli"
  sh "popper run -f {}"
}}}}
"""
    },
    # ##########################################################3
    "gitlab": {
        ".gitlab-ci.yml": """
---
image: docker:stable

variables:
  DOCKER_DRIVER: overlay

services:
- docker:dind

before_script:
- docker info
- apk update
- apk upgrade
- apk add python python-dev py-pip build-base git bash
- pip install virtualenv
- git clone https://github.com/systemslab/popper /tmp/popper
- pip install /tmp/popper/cli
popper:
  script: popper run -f {}
"""
    },
    # ##########################################################3
    "brigade": {
        "brigade.js": """
const { events, Job } = require("brigadier")
events.on("push", () => {
    var popper = new Job("popper", "python:3.7-slim-stretch")
    popper.tasks = [
        "apt-get update",
        "apt-get install -y git",
        "git clone https://github.com/systemslab/popper /tmp/popper",
        "export PYTHONUNBUFFERED=1",
        "pip install /tmp/popper/cli",
        "popper run -f {}"
    ]
    popper.run()
})
"""
    },
}


@click.command("ci", short_help="Generate CI service configuration files.")
@click.argument(
    "service",
    type=click.Choice(["travis", "circle", "jenkins", "gitlab", "brigade"]),
    required=True,
)
@click.option("-f", "--file", help="Specify workflow to run in CI.", required=True)
@pass_context
def cli(ctx, service, file):
    """Generates configuration files for distinct CI services. This command
    needs to be executed on the root of your Git repository folder.
    """
    if not os.path.exists(".git"):
        log.fail(
            "This command needs to be executed on the root of your "
            "Git project folder (where the .git/ folder is located)."
        )

    for ci_file, ci_file_content in ci_files[service].items():
        ci_file = os.path.join(os.getcwd(), ci_file)
        os.makedirs(os.path.dirname(ci_file), exist_ok=True)
        with open(ci_file, "w") as f:
            f.write(ci_file_content.format(file))

    log.info(f"Wrote {service} configuration successfully.")



import os
import unittest
import tempfile

from testfixtures import compare, Replacer, replace
from testfixtures.popen import MockPopen
from testfixtures.mock import call

from popper.config import ConfigLoader
from popper.runner import WorkflowRunner
from popper.parser import WorkflowParser
from popper.runner_slurm import (
    SlurmRunner,
    SingularityRunner,
)
from popper.cli import log as log

from .test_common import PopperTest

from box import Box


config = ConfigLoader.load(workspace_dir="/w")


def mock_kill(pid, sig):
    return 0


@unittest.skipIf(
    os.environ.get("ENABLE_SLURM_RUNNER_TESTS", "0") != "1",
    "Kubernetes runner tests not enabled.",
)
class TestSlurmSlurmRunner(PopperTest):
    def setUp(self):
        log.setLevel("CRITICAL")
        self.Popen = MockPopen()
        replacer = Replacer()
        replacer.replace("popper.runner_host.Popen", self.Popen)
        self.addCleanup(replacer.restore)

    def tearDown(self):
        log.setLevel("NOTSET")

    def test_tail_output(self):
        self.Popen.set_command("tail -f slurm-x.out", returncode=0)
        with SlurmRunner(config=ConfigLoader.load()) as sr:
            self.assertEqual(sr._tail_output("slurm-x.out"), 0)
            self.assertEqual(len(sr._out_stream_pid), 1)

    def test_stop_running_tasks(self):
        self.Popen.set_command("scancel --name job_a", returncode=0)
        with SlurmRunner(config=ConfigLoader.load()) as sr:
            sr._spawned_jobs.add("job_a")
            sr.stop_running_tasks()
            compare(
                call.Popen(
                    ["scancel", "--name", "job_a"],
                    cwd=os.getcwd(),
                    env=None,
                    preexec_fn=os.setsid,
                    stderr=-2,
                    stdout=-1,
                    universal_newlines=True,
                ),
                self.Popen.all_calls[0],
            )

    @replace("popper.runner_slurm.os.kill", mock_kill)
    def test_exec_srun(self, mock_kill):
        config_dict = {
            "engine": {"name": "singularity", "options": {},},
            "resource_manager": {
                "name": "slurm",
                "options": {"sample": {"gpus-per-task": 2, "overcommit": True}},
            },
        }

        config = ConfigLoader.load(workspace_dir="/w", config_file=config_dict)
        self.Popen.set_command(
            "srun --nodes 1 --ntasks 1 --ntasks-per-node 1 --gpus-per-task 2 --overcommit ls -la",
            returncode=0,
        )
        step = Box({"id": "sample"}, default_box=True)
        with SlurmRunner(config=config) as sr:
            e = sr._exec_srun(["ls", "-la"], step, logging=True)
            self.assertEqual(e, 0)

        call_srun = call.Popen(
            [
                "srun",
                "--nodes",
                "1",
                "--ntasks",
                "1",
                "--ntasks-per-node",
                "1",
                "--gpus-per-task",
                "2",
                "--overcommit",
                "ls",
                "-la",
            ],
            cwd=os.getcwd(),
            env=None,
            preexec_fn=os.setsid,
            stderr=-2,
            stdout=-1,
            universal_newlines=True,
        )

        self.assertEqual(call_srun in self.Popen.all_calls, True)

    @replace("popper.runner_slurm.os.kill", mock_kill)
    def test_exec_mpi(self, mock_kill):
        config_dict = {
            "engine": {"name": "singularity", "options": {},},
            "resource_manager": {
                "name": "slurm",
                "options": {"sample": {"gpus-per-task": 2, "overcommit": True}},
            },
        }

        config = ConfigLoader.load(workspace_dir="/w", config_file=config_dict)
        self.Popen.set_command(
            "sbatch "
            "--wait --gpus-per-task 2 --overcommit "
            f"popper_sample_{config.wid}.sh",
            returncode=0,
        )
        self.Popen.set_command(f"tail -f popper_sample_{config.wid}.out", returncode=0)
        step = Box({"id": "sample"}, default_box=True)
        with SlurmRunner(config=config) as sr:
            e = sr._exec_mpi(["ls -la"], step)
            self.assertEqual(e, 0)
            with open(f"popper_sample_{config.wid}.sh", "r") as f:
                content = f.read()

            self.assertEqual(
                content,
                f"""#!/bin/bash
#SBATCH --job-name=popper_sample_{config.wid}
#SBATCH --output=popper_sample_{config.wid}.out
#SBATCH --nodes=1
#SBATCH --ntasks=1
#SBATCH --ntasks-per-node=1
mpirun ls -la""",
            )
            self.assertEqual(len(sr._spawned_jobs), 0)
            self.assertEqual(sr._out_stream_thread.is_alive(), False)

        call_tail = call.Popen(
            ["tail", "-f", f"popper_sample_{config.wid}.out"],
            cwd=os.getcwd(),
            env=None,
            preexec_fn=os.setsid,
            stderr=-2,
            stdout=-1,
            universal_newlines=True,
        )

        call_sbatch = call.Popen(
            [
                "sbatch",
                "--wait",
                "--gpus-per-task",
                "2",
                "--overcommit",
                f"popper_sample_{config.wid}.sh",
            ],
            cwd=os.getcwd(),
            env=None,
            preexec_fn=os.setsid,
            stderr=-2,
            stdout=-1,
            universal_newlines=True,
        )

        self.assertEqual(call_tail in self.Popen.all_calls, True)
        self.assertEqual(call_sbatch in self.Popen.all_calls, True)

    def test_dry_run(self):
        config = ConfigLoader.load(
            engine_name="singularity", resman_name="slurm", dry_run=True
        )

        with WorkflowRunner(config) as r:
            wf_data = {
                "steps": [
                    {"uses": "docker://alpine", "runs": ["cat"], "args": ["README.md"],}
                ]
            }
            r.run(WorkflowParser.parse(wf_data=wf_data))

        self.assertEqual(self.Popen.all_calls, [])

    # @replace("popper.runner_slurm.os.kill", mock_kill)
    # def test_exec_srun_failure(self, mock_kill):
    #     config_dict = {
    #         "engine": {
    #             "name": "singularity",
    #             "options": {
    #                 "privileged": True,
    #                 "hostname": "popper.local",
    #                 "domainname": "www.example.org",
    #                 "volumes": ["/path/in/host:/path/in/container"],
    #                 "environment": {"FOO": "bar"},
    #             },
    #         },
    #         "resource_manager": {
    #             "name": "slurm",
    #             "options": {"1": {"nodes": 2, "nodelist": "worker1,worker2"}},
    #         },
    #     }

    #     config = ConfigLoader.load(workspace_dir="/w", config_file=config_dict)

    #     self.Popen.set_command(
    #         f"srun --nodes 2 --ntasks 2 --ntasks-per-node 1 --nodelist worker1,worker2 podman rm -f popper_1_{config.wid}",
    #         returncode=0,
    #     )

    #     self.Popen.set_command(
    #         f"srun --nodes 2 --ntasks 2 --ntasks-per-node 1 --nodelist worker1,worker2 podman pull alpine:latest",
    #         returncode=0,
    #     )

    #     self.Popen.set_command(
    #         f"srun --nodes 2 --ntasks 2 --ntasks-per-node 1 --nodelist worker1,worker2 podman create --name popper_1_{config.wid} --workdir /workspace -v /w:/workspace:Z -v /path/in/host:/path/in/container -e FOO=bar --privileged --hostname popper.local --domainname www.example.org alpine:latest ls",
    #         returncode=0,
    #     )

    #     self.Popen.set_command(
    #         f"srun --nodes 2 --ntasks 2 --ntasks-per-node 1 --nodelist worker1,worker2 podman start --attach popper_1_{config.wid}",
    #         returncode=12,
    #     )

    #     with WorkflowRunner(config) as r:
    #         wf_data = {"steps": [{"uses": "docker://alpine", "args": ["ls"]}]}
    #         self.assertRaises(SystemExit, r.run, WorkflowParser.parse(wf_data=wf_data))

    @replace("popper.runner_slurm.os.kill", mock_kill)
    def test_exec_mpi_failure(self, mock_kill):
        config_dict = {
            "engine": {"name": "singularity", "options": {},},
            "resource_manager": {
                "name": "slurm",
                "options": {
                    "1": {"nodes": 2, "nodelist": "worker1,worker2", "overcommit": True}
                },
            },
        }

        config = ConfigLoader.load(workspace_dir="/w", config_file=config_dict)

        self.Popen.set_command(
            "sbatch " "--wait --overcommit " f"popper_1_{config.wid}.sh", returncode=12,
        )

        self.Popen.set_command(f"tail -f popper_1_{config.wid}.out", returncode=0)

        with WorkflowRunner(config) as r:
            wf_data = {"steps": [{"uses": "docker://alpine", "args": ["ls"]}]}
            self.assertRaises(SystemExit, r.run, WorkflowParser.parse(wf_data=wf_data))


@unittest.skipIf(
    os.environ.get("ENABLE_SLURM_RUNNER_TESTS", "0") != "1",
    "Kubernetes runner tests not enabled.",
)
class TestSlurmSingularityRunner(unittest.TestCase):
    def setUp(self):
        self.Popen = MockPopen()
        replacer = Replacer()
        replacer.replace("popper.runner_host.Popen", self.Popen)
        self.addCleanup(replacer.restore)

    def tearDown(self):
        log.setLevel("NOTSET")

    def test_create_cmd(self):
        config = ConfigLoader.load(workspace_dir="/w")
        with SingularityRunner(config=config) as sr:
            step = Box({"args": ["-two", "-flags"]}, default_box=True)
            sr._container = "c1.sif"
            cmd = sr._create_cmd(step, "c1.sif")

            expected = (
                "singularity run"
                " --userns --pwd /workspace"
                " --bind /w:/workspace"
                " c1.sif"
                " -two -flags"
            )

            self.assertEqual(expected.split(" "), cmd)

        config_dict = {
            "engine": {
                "name": "singularity",
                "options": {
                    "hostname": "popper.local",
                    "ipc": True,
                    "bind": ["/path/in/host:/path/in/container"],
                },
            },
            "resource_manager": {"name": "slurm"},
        }

        config = ConfigLoader.load(workspace_dir="/w", config_file=config_dict)

        with SingularityRunner(config=config) as sr:
            step = Box({"args": ["-two", "-flags"]}, default_box=True)
            sr._container = "c2.sif"
            cmd = sr._create_cmd(step, "c2.sif")

            # fmt: off
            expected = f"singularity run --userns --pwd /workspace --bind /w:/workspace --bind /path/in/host:/path/in/container --hostname popper.local --ipc c2.sif -two -flags"
            # fmt: on
            self.assertEqual(expected.split(" "), cmd)

    @replace("popper.runner_slurm.os.kill", mock_kill)
    def test_run(self, mock_kill):
        self.maxDiff = None
        config_dict = {
            "engine": {
                "name": "singularity",
                "options": {
                    "hostname": "popper.local",
                    "bind": ["/path/in/host:/path/in/container"],
                },
            },
            "resource_manager": {
                "name": "slurm",
                "options": {
                    "1": {"nodes": 2, "ntasks": 2, "nodelist": "worker1,worker2"}
                },
            },
        }

        config = ConfigLoader.load(workspace_dir="/w", config_file=config_dict)

        # fmt: off
        self.Popen.set_command(
            "sbatch "
            "--wait "
            f"popper_1_{config.wid}.sh",
            returncode=0,
        )
        # fmt: on

        self.Popen.set_command(f"tail -f popper_1_{config.wid}.out", returncode=0)

        with WorkflowRunner(config) as r:
            wf_data = {"steps": [{"uses": "docker://alpine", "args": ["ls"]}]}
            r.run(WorkflowParser.parse(wf_data=wf_data))

        with open(f"popper_1_{config.wid}.sh", "r") as f:
            # fmt: off
            expected = f"""#!/bin/bash
#SBATCH --job-name=popper_1_{config.wid}
#SBATCH --output=popper_1_{config.wid}.out
#SBATCH --nodes=2
#SBATCH --ntasks=2
#SBATCH --ntasks-per-node=1
#SBATCH --nodelist=worker1,worker2
mpirun singularity run --userns --pwd /workspace --bind /w:/workspace --bind /path/in/host:/path/in/container --hostname popper.local {os.environ['HOME']}/.cache/popper/singularity/{config.wid}/popper_1_{config.wid}.sif ls"""
            # fmt: on
            actual = f.read()
            self.assertEqual(expected, actual)

        config_dict = {
            "engine": {
                "name": "singularity",
                "options": {
                    "hostname": "popper.local",
                    "bind": ["/path/in/host:/path/in/container"],
                },
            },
            "resource_manager": {
                "name": "slurm",
                "options": {
                    "1": {
                        "mpi": False,
                        "nodes": 2,
                        "ntasks": 2,
                        "nodelist": "worker1,worker2",
                    }
                },
            },
        }

        config = ConfigLoader.load(workspace_dir="/w", config_file=config_dict)

        self.Popen.set_command(
            f"srun --nodes 2 --ntasks 2 --ntasks-per-node 1 --nodelist worker1,worker2 rm -rf popper_1_{config.wid}.sif",
            returncode=0,
        )

        self.Popen.set_command(
            f"srun --nodes 2 --ntasks 2 --ntasks-per-node 1 --nodelist worker1,worker2 singularity pull docker://alpine:latest",
            returncode=0,
        )

        self.Popen.set_command(
            f"srun --nodes 2 --ntasks 2 --ntasks-per-node 1 --nodelist worker1,worker2 singularity run --userns --pwd /workspace --bind /w:/workspace --bind /path/in/host:/path/in/container --hostname popper.local {os.environ['HOME']}/.cache/popper/singularity/{config.wid}/popper_1_{config.wid}.sif ls",
            returncode=0,
        )

        with WorkflowRunner(config) as r:
            wf_data = {"steps": [{"uses": "docker://alpine", "args": ["ls"]}]}
            r.run(WorkflowParser.parse(wf_data=wf_data))



import os

from popper.config import ConfigLoader
from popper.cli import log

from .test_common import PopperTest

from box import Box


class TestPopperConfig(PopperTest):
    def setUp(self):
        log.setLevel("CRITICAL")
        self.maxDiff = None

    def tearDown(self):
        log.setLevel("NOTSET")

    def test_config_defaults(self):
        conf = ConfigLoader.load()
        expected = Box(
            {
                "skip_clone": False,
                "engine_name": "docker",
                "engine_opts": {},
                "resman_name": "host",
                "resman_opts": {},
                "skip_pull": False,
                "dry_run": False,
                "workspace_dir": os.getcwd(),
                "quiet": False,
                "reuse": False,
                "pty": False,
                "allow_undefined_secrets_in_ci": False,
            },
            default_box=True,
        )

        self.assertEqual(expected, TestPopperConfig.extract_dict(expected, conf))

    def test_config_non_defaults(self):
        expected = {
            "skip_clone": True,
            "skip_pull": True,
            "dry_run": True,
            "workspace_dir": os.path.realpath("/tmp/foo"),
            "quiet": True,
            "reuse": True,
            "pty": True,
            "allow_undefined_secrets_in_ci": True,
        }
        conf = ConfigLoader.load(**expected)
        self.assertEqual(expected, TestPopperConfig.extract_dict(expected, conf))

    def test_config_without_git_repo(self):
        conf = ConfigLoader.load(workspace_dir="/tmp/foo")
        self.assertTrue(not conf.git_commit)
        self.assertTrue(not conf.git_branch)
        self.assertTrue(not conf.git_sha_short)
        self.assertTrue(not conf.git_tag)

    def test_config_with_git_repo(self):
        r = self.mk_repo(tag="a-tag")
        conf = ConfigLoader.load(workspace_dir=r.working_dir)
        sha = r.head.object.hexsha
        self.assertEqual(r.git.rev_parse(sha), conf.git_commit)
        self.assertEqual(r.git.rev_parse(sha, short=7), conf.git_sha_short)
        self.assertEqual(r.active_branch.name, conf.git_branch)
        self.assertEqual(r.git.tag("--points-at", "HEAD"), conf.git_tag)

    def test_config_from_file(self):
        config = {
            "engine": {"options": {"privileged": True}},
            "resource_manager": {"options": {"foo": "bar"}},
        }
        kwargs = {"config_file": config}

        # engine name missing
        with self.assertLogs("popper", level="INFO") as cm:
            self.assertRaises(SystemExit, ConfigLoader.load, **kwargs)
            self.assertEqual(len(cm.output), 1)
            self.assertTrue("No engine name given" in cm.output[0])

        # resman name missing
        config.update({"engine": {"name": "foo"}})
        with self.assertLogs("popper", level="INFO") as cm:
            self.assertRaises(SystemExit, ConfigLoader.load, **kwargs)
            self.assertEqual(len(cm.output), 1)
            self.assertTrue("No resource manager name given" in cm.output[0])

        # now all OK
        config.update({"resource_manager": {"name": "bar"}})
        conf = ConfigLoader.load(**kwargs)
        self.assertEqual(conf.engine_name, "foo")
        self.assertEqual(conf.resman_name, "bar")
        self.assertEqual(conf.engine_opts, {})
        self.assertEqual(conf.resman_opts, {})

        config.update({"engine": {"name": "bar", "options": {"foo": "baz"}}})
        conf = ConfigLoader.load(**kwargs)
        self.assertEqual(conf.engine_opts, {"foo": "baz"})

    @staticmethod
    def extract_dict(A, B):
        # taken from https://stackoverflow.com/a/21213251
        return dict([(k, B[k]) for k in A.keys() if k in B.keys()])



import os
import unittest
import shutil

from unittest.mock import patch

import popper.scm as scm
from popper.cli import log
from popper.config import ConfigLoader
from popper.parser import WorkflowParser
from popper.runner import WorkflowRunner, StepRunner

from .test_common import PopperTest

from box import Box


class TestWorkflowRunner(unittest.TestCase):
    def setUp(self):
        log.setLevel("CRITICAL")

    def test_check_secrets(self):
        wf_data = {
            "steps": [
                {
                    "uses": "docker://alpine:3.9",
                    "args": ["ls", "-ltr"],
                    "secrets": ["SECRET_ONE", "SECRET_TWO"],
                }
            ]
        }
        wf = WorkflowParser.parse(wf_data=wf_data)

        # in dry-run, secrets are ignored
        runner = WorkflowRunner(ConfigLoader.load(dry_run=True))
        runner._process_secrets(wf)

        # now go back to not dry-running
        runner = WorkflowRunner(ConfigLoader.load())

        # when CI=true it should fail
        os.environ["CI"] = "true"
        self.assertRaises(SystemExit, runner._process_secrets, wf)

        # but it should be fine if we allow undefined secrets
        runner = WorkflowRunner(ConfigLoader.load(allow_undefined_secrets_in_ci=True))
        runner._process_secrets(wf)

        # add one secret
        os.environ["SECRET_ONE"] = "1234"

        # it should fail, as we're missing one
        runner = WorkflowRunner(ConfigLoader.load())
        self.assertRaises(SystemExit, runner._process_secrets, wf)

        os.environ.pop("CI")

        # now is fine
        with patch("getpass.getpass", return_value="5678"):
            runner._process_secrets(wf)

        # pop the other
        os.environ.pop("SECRET_ONE")

    def test_clone_repos(self):
        wf_data = {"steps": [{"uses": "popperized/bin/sh@master"}]}
        wf = WorkflowParser.parse(wf_data=wf_data)

        cache_dir = os.path.join(os.environ["HOME"], ".cache/popper/")

        # clone repos in the default cache directory.
        conf = ConfigLoader.load()
        runner = WorkflowRunner(conf)
        runner._clone_repos(wf)
        step_dir = os.path.join(cache_dir, conf.wid, "github.com/popperized/bin")
        self.assertTrue(os.path.exists(step_dir))

        # clone repos in custom cache directory
        os.environ["POPPER_CACHE_DIR"] = "/tmp/smdir"
        conf = ConfigLoader.load()
        runner = WorkflowRunner(conf)
        runner._clone_repos(wf)
        step_dir = os.path.join("/tmp/smdir", conf.wid, "github.com/popperized/bin")
        self.assertTrue(os.path.exists(step_dir))
        os.environ.pop("POPPER_CACHE_DIR")

        # check failure when container is not available and we skip cloning
        shutil.rmtree("/tmp/smdir")
        shutil.rmtree(cache_dir)
        conf = ConfigLoader.load(skip_clone=True)
        runner = WorkflowRunner(conf)
        self.assertRaises(SystemExit, runner._clone_repos, wf)

    def test_steprunner_factory(self):
        with WorkflowRunner(ConfigLoader.load()) as r:
            self.assertEqual(
                "HostRunner", r._step_runner("host", None).__class__.__name__
            )


class TestStepRunner(PopperTest):
    def setUp(self):
        log.setLevel("CRITICAL")

    def test_prepare_environment_without_git(self):
        step = Box(
            {"name": "a", "env": {"FOO": "BAR"}, "secrets": ["A"]}, default_box=True
        )
        os.environ["A"] = "BC"

        with StepRunner(ConfigLoader.load(workspace_dir="/tmp/foo")) as r:
            env = r._prepare_environment(step, {"other": "b"})
            self.assertDictEqual({"FOO": "BAR", "A": "BC", "other": "b"}, env)
            os.environ.pop("A")

            # secret undefined should return an empty variable
            env = r._prepare_environment(step, {"other": "b"})
            self.assertDictEqual({"FOO": "BAR", "A": "", "other": "b"}, env)

    def test_prepare_environment_with_git(self):
        repo = self.mk_repo(tag="a-tag")
        conf = ConfigLoader.load(workspace_dir=repo.working_dir)
        with StepRunner(conf) as r:
            step = Box(
                {"name": "a", "env": {"FOO": "BAR"}, "secrets": ["A"]}, default_box=True
            )
            os.environ["A"] = "BC"
            env = r._prepare_environment(step, {"other": "b"})
            expected = {
                "FOO": "BAR",
                "A": "BC",
                "other": "b",
                "GIT_COMMIT": conf.git_commit,
                "GIT_BRANCH": conf.git_branch,
                "GIT_SHA_SHORT": conf.git_sha_short,
                "GIT_REMOTE_ORIGIN_URL": conf.git_remote_origin_url,
                "GIT_TAG": conf.git_tag,
            }
            self.assertDictEqual(expected, env)
            os.environ.pop("A")

    def test_get_build_info(self):
        step = Box(
            {"uses": "popperized/bin/sh@master", "args": ["ls"], "id": "one",},
            default_box=True,
        )
        with StepRunner() as r:
            build, _, img, tag, build_ctx_path = r._get_build_info(step)
            self.assertEqual(build, True)
            self.assertEqual(img, "popperized/bin")
            self.assertEqual(tag, "master")
            self.assertTrue(f"{os.environ['HOME']}/.cache/popper" in build_ctx_path)
            self.assertTrue("github.com/popperized/bin/sh" in build_ctx_path)

        step = Box(
            {
                "uses": "docker://alpine:3.9",
                "runs": ["sh", "-c", "echo $FOO > hello.txt ; pwd"],
                "env": {"FOO": "bar"},
                "id": "1",
            },
            default_box=True,
        )
        with StepRunner() as r:
            build, _, img, tag, build_sources = r._get_build_info(step)
            self.assertEqual(build, False)
            self.assertEqual(img, "alpine")
            self.assertEqual(tag, "3.9")
            self.assertEqual(build_sources, None)

        step = Box({"uses": "./", "args": ["ls"], "id": "one",}, default_box=True,)
        conf = ConfigLoader.load(workspace_dir="/tmp")
        with StepRunner(config=conf) as r:
            build, _, img, tag, build_ctx_path = r._get_build_info(step)
            self.assertEqual(build, True)
            self.assertEqual(img, "popper_one_step")
            self.assertEqual(tag, "na")
            self.assertEqual(build_ctx_path, f"{os.path.realpath('/tmp')}/./")

        # test within a git repo
        repo = self.mk_repo()
        conf = ConfigLoader.load(workspace_dir=repo.working_dir)
        with StepRunner(config=conf) as r:
            build, _, img, tag, build_ctx_path = r._get_build_info(step)
            self.assertEqual(build, True)
            self.assertEqual(img, "popper_one_step")
            self.assertEqual(tag, scm.get_sha(repo, short=7))
            self.assertEqual(build_ctx_path, f"{os.path.realpath(repo.working_dir)}/./")

        step = Box(
            {
                "uses": "docker://alpine:3.9",
                "runs": ["sh", "-c", "echo $FOO > hello.txt ; pwd"],
                "env": {"FOO": "bar"},
                "name": "1",
            },
            default_box=True,
        )
        with StepRunner() as r:
            build, img_full, _, _, build_ctx_path = r._get_build_info(step)
            self.assertEqual(build, False)
            self.assertEqual(img_full, "docker://alpine:3.9")
            self.assertEqual(build_ctx_path, None)

    def test_get_container_kwargs(self):
        step = Box(
            {
                "uses": "popperized/bin/sh@master",
                "args": ["ls"],
                "id": "one",
                "dir": "/tmp/",
                "options": {"ports": {"8888/tcp": 8888}},
            },
            default_box=True,
        )

        config_dict = {
            "engine": {
                "name": "docker",
                "options": {
                    "privileged": True,
                    "hostname": "popper.local",
                    "domainname": "www.example.org",
                    "volumes": ["/path/in/host:/path/in/container"],
                    "environment": {"FOO": "bar"},
                },
            },
        }

        config = ConfigLoader.load(
            config_file=config_dict, workspace_dir="/path/to/workdir"
        )

        with StepRunner(config=config) as r:
            args = r._get_container_kwargs(step, "alpine:3.9", "container_a")
            self.assertEqual(
                args,
                {
                    "image": "alpine:3.9",
                    "command": ["ls"],
                    "name": "container_a",
                    "volumes": [
                        "/path/to/workdir:/workspace:Z",
                        "/path/in/host:/path/in/container",
                    ],
                    "working_dir": "/tmp/",
                    "environment": {"FOO": "bar"},
                    "entrypoint": None,
                    "detach": True,
                    "stdin_open": False,
                    "tty": False,
                    "privileged": True,
                    "hostname": "popper.local",
                    "domainname": "www.example.org",
                    "ports": {"8888/tcp": 8888},
                },
            )

        # check container kwargs when pty is enabled
        config = ConfigLoader.load(
            config_file=config_dict, workspace_dir="/path/to/workdir", pty=True
        )

        with StepRunner(config=config) as r:
            args = r._get_container_kwargs(step, "alpine:3.9", "container_a")

            self.assertEqual(
                args,
                {
                    "image": "alpine:3.9",
                    "command": ["ls"],
                    "name": "container_a",
                    "volumes": [
                        "/path/to/workdir:/workspace:Z",
                        "/path/in/host:/path/in/container",
                    ],
                    "working_dir": "/tmp/",
                    "environment": {"FOO": "bar"},
                    "entrypoint": None,
                    "detach": False,
                    "stdin_open": True,
                    "tty": True,
                    "privileged": True,
                    "hostname": "popper.local",
                    "domainname": "www.example.org",
                    "ports": {"8888/tcp": 8888},
                },
            )



import getpass
import importlib
import os
import sys

import popper.scm as scm
import popper.utils as pu

from popper.config import ConfigLoader
from popper.cli import log


class WorkflowRunner(object):
    """The workflow runner."""

    # class variable that holds references to runner singletons
    __runners = {}

    def __init__(self, config):
        self._config = config
        self._is_resman_module_loaded = False

    def _load_resman_module(self):
        """dynamically load resource manager module."""
        resman_mod_name = f"popper.runner_{self._config.resman_name}"
        resman_spec = importlib.util.find_spec(resman_mod_name)
        if not resman_spec:
            raise ValueError(f"Invalid resource manager: {self._config.resman_name}")
        self._resman_mod = importlib.import_module(resman_mod_name)
        self._is_resman_module_loaded = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, traceback):
        """calls __exit__ on all instantiated step runners"""
        if self._config.repo:
            self._config.repo.close()
        for _, r in WorkflowRunner.__runners.items():
            r.__exit__(exc_type, exc, traceback)
        WorkflowRunner.__runners = {}

    @staticmethod
    def signal_handler(sig, frame):
        """Handles the interrupt signal.

        Args:
            sig(int): Signal number of signal being passed to cli.
            frame(class):It represents execution frame. For more see
                https://docs.python.org/3/reference/datamodel.html

        Returns:
            None
        """
        log.info(f"Got {sig} signal. Stopping running steps.")
        for _, runner in WorkflowRunner.__runners.items():
            runner.stop_running_tasks()
        sys.exit(0)

    def _process_secrets(self, wf):
        """Checks whether the secrets defined for a step are available in the
        execution environment. When the environment variable `CI` is set to
        `true` and no environment variable is defined for a secret, the
        execution fails, otherwise (CI not set or CI=false), it prompts the
        user to enter the values for undefined secrets. The processing is
        completely skipped if config.dry_run=True.

        Args:
          wf(popper.parser.Workflow): Instance of the Workflow class.
          config.dry_run(bool): skip execution of this function
          config.skip_clone(bool): skip execution of this function
        Returns:
            None
        """
        if self._config.dry_run or self._config.skip_clone:
            return

        for step in wf.steps:
            for s in step.secrets:
                if s not in os.environ:
                    if not os.environ.get("CI", None):
                        val = getpass.getpass(f"Enter the value for {s} : ")
                        os.environ[s] = val
                    elif not self._config.allow_undefined_secrets_in_ci:
                        log.fail(f"Secret {s} not defined")

    def _clone_repos(self, wf):
        """Clone steps that reference a repository.

        Args:
          wf(popper.parser.workflow): Instance of the Workflow class.
          config.dry_run(bool): True if workflow flag is being dry-run.
          config.skip_clone(bool): True if clonning step has to be skipped.
          config.wid(str): id of the workspace

        Returns:
            None
        """
        # cache directory for this workspace
        wf_cache_dir = os.path.join(self._config.cache_dir, self._config.wid)
        os.makedirs(wf_cache_dir, exist_ok=True)

        cloned = set()
        infoed = False

        for step in wf.steps:
            if (
                "docker://" in step.uses
                or "shub://" in step.uses
                or "library://" in step.uses
                or "./" in step.uses
                or step.uses == "sh"
            ):
                continue

            url, service, user, repo, _, version = scm.parse(step.uses)

            repo_dir = os.path.join(wf_cache_dir, service, user, repo)

            if self._config.dry_run:
                continue

            if self._config.skip_clone:
                if not os.path.exists(repo_dir):
                    log.fail(f"Expecting folder '{repo_dir}' not found.")
                continue

            if not infoed:
                log.info("[popper] Cloning step repositories")
                infoed = True

            if f"{user}/{repo}" in cloned:
                continue

            log.info(f"[popper] - {url}/{user}/{repo}@{version}")
            scm.clone(url, user, repo, repo_dir, version)
            cloned.add(f"{user}/{repo}")

    def run(self, wf):
        """Run the given workflow.

        Args:
          wf(Workflow): workflow to be executed

        Returns:
            None
        """
        self._process_secrets(wf)
        self._clone_repos(wf)

        for step in wf.steps:
            log.debug(f"Executing step:\n{pu.prettystr(step)}")
            if step.uses == "sh":
                e = self._step_runner("host", step).run(step)
            else:
                e = self._step_runner(self._config.engine_name, step).run(step)

            if e != 0 and e != 78:
                log.fail(f"Step '{step.id}' failed ('{e}') !")

            log.info(f"Step '{step.id}' ran successfully !")

            if e == 78:
                break

        log.info("Workflow finished successfully.")

    def _step_runner(self, engine_name, step):
        """Factory of singleton runners."""
        if not self._is_resman_module_loaded:
            self._load_resman_module()

        runner = WorkflowRunner.__runners.get(engine_name, None)

        if not runner:
            engine_cls_name = f"{engine_name.capitalize()}Runner"
            engine_cls = getattr(self._resman_mod, engine_cls_name, None)
            if not engine_cls:
                raise ValueError(f"Cannot find class for {engine_name}")
            runner = engine_cls(config=self._config)
            WorkflowRunner.__runners[engine_name] = runner

        return runner


# class design guidelines:
# - if not exposed to users, then it is protected, e.g `_foo()`
# - if a method does not use internal state then it is a @staticmethod
# - if both of the above, then it's both protected and static
class StepRunner(object):
    """Base class for step runners, assumed to be singletons."""

    def __init__(self, config=None):
        if not config:
            self._config = ConfigLoader.load()
        else:
            self._config = config

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, traceback):
        pass

    def _prepare_environment(self, step, env={}):
        """Prepare environment variables for a step, which includes those in
        the 'env' and 'secrets' attributes.

        Args:
          step(dict): step information
          env(dict): optional environment to include in returned environment

        Returns:
          dict: key-value map of environment variables.
        """
        step_env = step.env.to_dict()
        for s in step.secrets:
            step_env.update({s: os.environ.get(s, "")})
        step_env.update(env)

        # define GIT_* variables
        if self._config.repo:
            step_env.update(
                {
                    "GIT_COMMIT": self._config.git_commit,
                    "GIT_BRANCH": self._config.git_branch,
                    "GIT_SHA_SHORT": self._config.git_sha_short,
                    "GIT_REMOTE_ORIGIN_URL": self._config.git_remote_origin_url,
                    # git_tag might be None, so we use empty string instead
                    "GIT_TAG": self._config.get("git_tag", ""),
                }
            )
        return step_env

    def _get_build_info(self, step):

        """Parses the `uses` attribute and returns build information needed.

            Args:
                step(dict): dict with step data
            Returns:
                (str, str, str, str): bool (build), image, tag, Dockerfile
            """
        build = True
        img = None
        build_ctx_path = None
        img_full = None
        tag = None

        if "docker://" in step.uses:
            img_full = step.uses
            img = step.uses.replace("docker://", "")
            if ":" in img:
                (img, tag) = img.split(":")
            else:
                tag = "latest"
            build = False

        elif "./" in step.uses:
            img_full = f'{pu.sanitized_name(step.id, "step")}'
            img = img_full.lower()
            tag = self._config.git_sha_short if self._config.git_sha_short else "na"
            build_ctx_path = os.path.join(self._config.workspace_dir, step.uses)

        else:
            _, service, user, repo, step_dir, version = scm.parse(step.uses)
            wf_cache_dir = os.path.join(self._config.cache_dir, self._config.wid)
            repo_dir = os.path.join(wf_cache_dir, service, user, repo)
            img_full = f"{user}/{repo}".lower()
            img = img_full
            tag = version
            build_ctx_path = os.path.join(repo_dir, step_dir)

        return (build, img_full, img, tag, build_ctx_path)

    def _update_with_engine_config(self, container_args):
        """Given container arguments, it extends it so it includes options
        obtained from the popper.config.Config.engine_opts property.
        """
        update_with = self._config.engine_opts
        if not update_with:
            return

        if container_args.get("volumes"):
            container_args["volumes"] = [
                *container_args["volumes"],
                *update_with.get("volumes", list()),
            ]
        if container_args.get("bind"):
            container_args["bind"] = [
                *container_args["bind"],
                *update_with.get("bind", list()),
            ]

        for k, v in update_with.get("environment", dict()).items():
            container_args["environment"].update({k: v})

        for k, v in update_with.items():
            if k not in container_args.keys():
                container_args[k] = update_with[k]

    def _get_container_kwargs(self, step, img, name):
        args = {
            "image": img,
            "command": list(step.args),
            "name": name,
            "volumes": [f"{self._config.workspace_dir}:/workspace:Z",],
            "working_dir": step.dir if step.dir else "/workspace",
            "environment": self._prepare_environment(step),
            "entrypoint": step.runs if step.runs else None,
            "detach": not self._config.pty,
            "tty": self._config.pty,
            "stdin_open": self._config.pty,
        }

        self._update_with_engine_config(args)
        args.update(step.options)
        log.debug(f"container args: {pu.prettystr(args)}\n")

        return args

    def stop_running_tasks(self):
        raise NotImplementedError("Needs implementation in derived classes.")

    def run(self, step):
        raise NotImplementedError("Needs implementation in derived classes.")



import os
import tempfile
import git
import unittest


class PopperTest(unittest.TestCase):
    def mk_repo(self, tag=None):
        """creates a test repo in a random temp file. Equivalent to:
        REPODIR=/tmp/<random>
        mkdir $REPODIR
        cd $REPODIR
        git init
        touch README.md
        git add .
        git commit -m 'first commit'
        echo 'README content' > README.md
        git add -u .
        git commit -m 'second commit'
        git tag <tag>
        """
        tempdir = tempfile.mkdtemp()
        repo = git.Repo.init(tempdir)

        # touch README file
        readme = os.path.join(tempdir, "README.md")
        open(readme, "w").close()

        # create first commit
        repo.index.add([readme])
        repo.index.commit("first commit")

        # modify README file and create second commit
        with open(readme, "w") as f:
            f.write("README content\n")
        repo.index.add([readme])
        repo.index.commit("second commit")

        # tag commit
        if tag:
            repo.create_tag(tag)

        # create a remote
        repo.create_remote("origin", url="https://github.com/my/repo")

        return repo


class PopperCommonTest(PopperTest):
    def test_mkrepo(self):

        repo = self.mk_repo()
        self.assertTrue(os.path.isdir(repo.working_tree_dir))
        self.assertTrue(
            os.path.isfile(os.path.join(repo.working_tree_dir, "README.md"))
        )



import os
import yaml

import popper.scm as scm

from hashlib import shake_256
from popper.cli import log as log

from box import Box


class ConfigLoader(object):
    @staticmethod
    def load(
        engine_name=None,
        resman_name=None,
        config_file=None,
        workspace_dir=os.getcwd(),
        reuse=False,
        dry_run=False,
        quiet=False,
        skip_pull=False,
        skip_clone=False,
        pty=False,
        allow_undefined_secrets_in_ci=False,
    ):
        """Loads and creates a configuration, represented by a frozen Box
        """
        workspace_dir = os.path.realpath(workspace_dir)
        repo = scm.new_repo(workspace_dir)

        # path to cache
        if os.environ.get("POPPER_CACHE_DIR", None):
            cache_dir = os.environ["POPPER_CACHE_DIR"]
        else:
            cache_dir_default = os.path.join(os.environ["HOME"], ".cache")
            cache_dir = os.environ.get("XDG_CACHE_HOME", cache_dir_default)
            cache_dir = os.path.join(cache_dir, "popper")

        from_file = ConfigLoader.__load_config_from_file(
            config_file, engine_name, resman_name
        )

        pp_config = {
            "workspace_dir": workspace_dir,
            "reuse": reuse,
            "dry_run": dry_run,
            "quiet": quiet,
            "skip_pull": skip_pull,
            "skip_clone": skip_clone,
            "pty": pty,
            "allow_undefined_secrets_in_ci": allow_undefined_secrets_in_ci,
            # if no git repository exists in workspace_dir or its parents, the repo
            # variable is None and all git_* variables are assigned to 'na'
            "repo": repo,
            "git_commit": scm.get_sha(repo),
            "git_sha_short": scm.get_sha(repo, short=7),
            "git_branch": scm.get_branch(repo),
            "git_tag": scm.get_tag(repo),
            "git_remote_origin_url": scm.get_remote_url(repo),
            # wid is used to associate a unique id to this workspace. This is then
            # used by runners to name resources in a way that there is no name
            # clash between concurrent workflows being executed
            "wid": shake_256(workspace_dir.encode("utf-8")).hexdigest(4),
            "cache_dir": cache_dir,
            "engine_name": from_file["engine_name"],
            "resman_name": from_file["resman_name"],
            "engine_opts": from_file["engine_opts"],
            "resman_opts": from_file["resman_opts"],
        }

        return Box(pp_config, default_box=True, frozen_box=True)

    @staticmethod
    def __load_config_from_file(config_file, engine_name, resman_name):
        from_file = ConfigLoader.__load_config_file(config_file)
        loaded_conf = {}

        eng_section = from_file.get("engine", None)
        eng_from_file = from_file.get("engine", {}).get("name")
        if from_file and eng_section and not eng_from_file:
            log.fail("No engine name given.")

        resman_section = from_file.get("resource_manager", None)
        resman_from_file = from_file.get("resource_manager", {}).get("name")
        if from_file and resman_section and not resman_from_file:
            log.fail("No resource manager name given.")

        # set name in precedence order (or assigne default values)
        if engine_name:
            loaded_conf["engine_name"] = engine_name
        elif eng_from_file:
            loaded_conf["engine_name"] = eng_from_file
        else:
            loaded_conf["engine_name"] = "docker"

        if resman_name:
            loaded_conf["resman_name"] = resman_name
        elif resman_from_file:
            loaded_conf["resman_name"] = resman_from_file
        else:
            loaded_conf["resman_name"] = "host"

        engine_opts = from_file.get("engine", {}).get("options", {})
        resman_opts = from_file.get("resource_manager", {}).get("options", {})
        loaded_conf["engine_opts"] = engine_opts
        loaded_conf["resman_opts"] = resman_opts

        return loaded_conf

    @staticmethod
    def __load_config_file(config_file):
        """Validate and parse the engine configuration file.

        Args:
          config_file(str): Path to the file to be parsed.

        Returns:
          dict: Engine configuration.
        """
        if isinstance(config_file, dict):
            return config_file

        if not config_file:
            return dict()

        if not os.path.exists(config_file):
            log.fail(f"File {config_file} was not found.")

        if not config_file.endswith(".yml"):
            log.fail("Configuration file must be a YAML file.")

        with open(config_file, "r") as cf:
            data = yaml.load(cf, Loader=yaml.Loader)

        if not data:
            log.fail("Configuration file is empty.")

        return data



import click

from popper import __version__ as popper_version
from popper.cli import pass_context, log


@click.command("version", short_help="Show version of Popper and exit.")
@pass_context
def cli(ctx):
    """Displays version of Popper and exit.
    """

    # Args:
    #   ctx(Popper.cli.context): For process inter-command communication
    #         context is used.For reference visit
    #         https://click.palletsprojects.com/en/7.x/commands/#nested-handling-and-contexts .

    # Returns:
    #   None

    log.info(f"Popper version {popper_version}")



import tempfile

from click.testing import CliRunner

import popper.commands.cmd_dot as dot

from popper.cli import log

from .test_common import PopperTest


class TestDot(PopperTest):
    def setUp(self):
        log.setLevel("CRITICAL")

    def test_dot(self):
        with tempfile.NamedTemporaryFile(mode="w+t", suffix=".yml") as f:
            f.write(
                """
steps:
- id: one
  uses: 'foo/bar'
- id: two
  uses: 'foo/bar'
- id: three
  uses: 'foo/bar'
"""
            )
            f.flush()

            runner = CliRunner()

            with self.assertLogs("popper") as test_logger:
                result = runner.invoke(dot.cli, ["-f", f.name])
                log.debug(f"file: {f.name}")
                self.assertEqual(result.exit_code, 0)
                log_output = test_logger.output[0]
                self.assertTrue('"one" -> "two";' in log_output)
                self.assertTrue('"two" -> "three";' in log_output)
                self.assertFalse("fillcolor=transparent,color=" in log_output)

            with self.assertLogs("popper") as test_logger:

                result = runner.invoke(dot.cli, ["--colors", "-f", f.name])
                self.assertEqual(result.exit_code, 0)
                self.assertTrue("fillcolor=transparent,color=" in test_logger.output[0])



import click

from popper.cli import pass_context, log, PopperCLI

popper_cli = PopperCLI()


@click.command("help", short_help="Shows help for a given command.")
@click.argument(
    "subcommand",
    type=click.Choice(popper_cli.list_commands(click.Context(popper_cli))),
    required=False,
)
@pass_context
def cli(ctx, subcommand):
    """ Display help for a given command or popper default help
    """
    if subcommand:
        target_command = popper_cli.get_command(ctx, subcommand)
        log.info(target_command.get_help(click.Context(popper_cli)))
    else:
        log.info(popper_cli.get_help(click.Context(popper_cli)))



import click

from popper.cli import pass_context, log
from popper.parser import WorkflowParser


@click.option(
    "-f",
    "--file",
    help="File containing the definition of the workflow.",
    required=True,
)
@click.option(
    "--skip",
    help=("Skip the list of steps specified."),
    required=False,
    default=list(),
    multiple=True,
)
@click.option(
    "--colors", help="Use colors in the graph.", required=False, is_flag=True,
)
@click.command("dot", short_help="Generate a graph in the .dot format.")
@pass_context
def cli(ctx, file, skip, colors):
    """Creates a graph in the .dot format representing the workflow."""
    wf = WorkflowParser.parse(file=file)

    node_attrs = 'shape=box, style="filled{}", fillcolor=transparent{}'
    wf_attr = node_attrs.format(",rounded", ",color=red" if colors else "")
    act_attr = node_attrs.format("", ",color=cyan" if colors else "")
    dot_str = ""
    dot_str += f'  "Workflow" [{wf_attr}];\n'
    for i, step in enumerate(wf.steps):
        n = wf.steps[i]["id"]
        dot_str += f'  "{n}" [{act_attr}];\n'
        if i == 0:
            parent = "Workflow"
        else:
            parent = wf.steps[i - 1]["id"]
        dot_str += f'  "{parent}" -> "{n}";\n'
    log.info("digraph G { graph [bgcolor=transparent];\n" + dot_str + "}\n")
